# -*- coding: utf-8 -*-
# @Time    : 2023/3/31 22:18
# @Author  : luyi
from .constants import *
from .lineExpr import LineExpr
from .model import Model
from .cpmodel import CpModel
from .tupledict import tupledict, multidict
from .tuplelist import tuplelist
from .utilx import name_str
from .func import debugVar
from .variable import Var, IntervalVar
from ._version import __version__

BINARY = Vtype.BINARY
CONTINUOUS = Vtype.CONTINUOUS
INTEGER = Vtype.INTEGER
MINIMIZE, MAXIMIZE = ObjType
LESS_EQUAL, EQUAL, GREATER_EQUAL = CmpType
# 最优解和可行解
OPTIMAL, FEASIBLE = OptimizationStatus.OPTIMAL, OptimizationStatus.FEASIBLE
name = "mipx"



# -*- coding: utf-8 -*-
# @Time    : 2023/3/31 22:21
# @Author  : luyi
from enum import Enum
from typing import Literal, Optional


P1, P2, P3, P4, P5, P6, P7, P8, P9 = (
    1,
    10,
    100,
    1000,
    10000,
    100000,
    1000000,
    10000000,
    100000000,
)


class Vtype(Enum):
    BINARY = "B"
    CONTINUOUS = "C"
    INTEGER = "I"


class ObjType(Enum):
    MINIMIZE = "MIN"
    MAXIMIZE = "MAX"


# optimization status
class OptimizationStatus(Enum):
    """求解器返回错误"""
    ERROR = -1

    """已计算出最优解"""
    OPTIMAL = 0

    """模型被证明无可行解"""
    INFEASIBLE = 1

    """目标函数中存在一个或多个变量未受约束条件限制，且最优目标值为无穷大"""
    UNBOUNDED = 2

    """在搜索过程中找到了整数可行解，但在确认该解是否为最优解前搜索被中断"""
    FEASIBLE = 3

    """松弛线性规划存在可行解，但带整数变量的问题无可行解"""
    INT_INFEASIBLE = 4

    """执行了截断搜索但未找到任何整数可行解"""
    NO_SOLUTION_FOUND = 5

    """问题已加载但未执行优化"""
    LOADED = 6

    """当前截断值下不存在可行解"""
    CUTOFF = 7

    """其他状态"""
    OTHER = 10000


class CmpType(Enum):
    LESS_EQUAL = 0
    EQUAL = 1
    GREATER_EQUAL = 2


class Params:
    def __init__(self):
        self.TimeLimit: Optional[int] = None  # 单位秒
        self.MIPGap: Optional[float] = None  # 表示多少的gap,0.10则表示10
        self.EnableOutput = False
        self.Precision: Optional[float] = None  # 精度控制
        self.CpoptimizerPath: Optional[str] = None  # Cplex约束求解器路径



# 生成测试
import unittest
import mipx


class TestUtils(unittest.TestCase):

    def test_debug_var(self):
        model = mipx.Model(solver_id="CBC")
        x = model.addVars(1, 3, name="x")
        y = model.addVar(name="y")
        status = model.optimize()
        if status == mipx.OptimizationStatus.OPTIMAL:
            mipx.debugVar(x.select("*", 2), False)
            mipx.debugVar(y, False)
            mipx.debugVar(x, False)
            mipx.debugVar([y], False)

    def test_tuple_list(self):
        a = mipx.tuplelist([1, 2, 3])
        print(a.quickselect(1))

    def test_cp_model(self):
        pass


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# @Time    : 2023/4/29 10:13
# @Author  : luyi
import numbers

try:
    from ortools.linear_solver.linear_solver_natural_api import CastToLinExp, OFFSET_KEY  # type: ignore
except Exception as e:
    from ortools.linear_solver.python.linear_solver_natural_api import (
        CastToLinExp,
        OFFSET_KEY,
    )

from ortools.linear_solver.pywraplp import SumArray
from .utilx import is_list_or_tuple
from .interface_ import IVar


def SumExpr(*args):
    return LineExpr(args)


class LineExpr(SumArray):
    def __init__(self, array=None):
        if array is None:
            array = []
        else:
            if not is_list_or_tuple(array):
                array = [array]
        super().__init__(array)
        self._array: list = self._SumArray__array  # type: ignore

    def add(self, arg, mult=1.0):
        self._array.append(CastToLinExp(arg * mult))

    def addConstant(self, constant):
        """
        添加常量
        :param constant:
        :return:
        """
        self._array.append(CastToLinExp(constant))

    def addTerms(self, newcoeffs, newvars):
        """
        通过系数和变量来添加线性表达式
            addTerms(1.0, x)
            addTerms([1.0, 2.0], [x, y])
        :param newcoeffs: 系数集合或者系数
        :param newvars: 变量集合或者变量
        :return:
        """
        if is_list_or_tuple(newcoeffs) == is_list_or_tuple(newvars):
            if is_list_or_tuple(newcoeffs) and len(newcoeffs) != len(newvars):
                raise TypeError("传参异常")
            # 动态添加
            if is_list_or_tuple(newcoeffs):
                for i in range(len(newcoeffs)):
                    self.add(newvars[i], newcoeffs[i])
            else:
                self.add(newvars, newcoeffs)
        else:
            raise TypeError("传参异常")

    def clear(self):
        self._array.clear()

    def copy(self):
        return LineExpr(self._array)

    def getCoeff(self, x: IVar):
        """
        获取第i个系数
        :param x:
        :return:
        """
        coeffs = self.GetCoeffs()
        return coeffs.get(x)

    def getConstant(self):
        coeffs = self.GetCoeffs()
        KEY = OFFSET_KEY
        print(coeffs)
        if KEY in coeffs:
            return coeffs.get(KEY)
        else:
            return 0

    def getValue(self) -> numbers.Number:
        """
        表达式的值
        :return:
        """
        value = 0
        coeffs = self.GetCoeffs()
        for key, coeff in coeffs.items():
            if key is OFFSET_KEY:
                value += coeffs.get(key)  # type: ignore
            else:
                var: IVar = key
                x = var.X
                value += coeff * x
        return value  # type: ignore

    def size(self):
        return len(self.GetCoeffs())

    def __add__(self, expr):
        return SumExpr(self, expr)

    def __radd__(self, cst):
        return SumExpr(self, cst)

    def __sub__(self, expr):
        return SumExpr(self, -expr)

    def __rsub__(self, cst):
        return SumExpr(-self, cst)

    def __mul__(self, other):  # type: ignore
        pass



import mipx


if __name__ == '__main__':
    model = mipx.Model()
    x = model.addVars(2, vtype=mipx.BINARY, name="x")
    z = model.addVar(vtype=mipx.BINARY, name="z")
    model.addGenConstrOr(z, [x[0], x[1]])
    model.addConstr(x[0]+x[1] == 1)

    y1 = model.addVar(vtype=mipx.INTEGER, name="y")
    y2 = model.addVar(vtype=mipx.INTEGER, name="y")
    model.addConstr(y1-y2 == -4)
    z1 = model.addVar(vtype=mipx.INTEGER, name="z")
    s1 = model.addVar(vtype=mipx.INTEGER, name="s1")
    e1 = model.addVar(vtype=mipx.INTEGER, name="e1")
    s2 = model.addVar(vtype=mipx.INTEGER, name="s2")
    e2 = model.addVar(vtype=mipx.INTEGER, name="e2")
    ivar = model.newIntervalVar(s1, 10, e1, name="interval")
    ivar2 = model.newIntervalVar(s2, 10, e2, name="interval2")
    model.addNoOverlap([ivar, ivar2], M=100)

    model.addGenConstrAbs(z1, y1-y2, M=10000)
    status = model.optimize()
    if status == mipx.OptimizationStatus.OPTIMAL:
        print(f"Optimal value: {model.ObjVal}")
        print(f"z = {z.X}")
        print(z1.X)
        mipx.debugVar(s1)
        mipx.debugVar(e1)
        mipx.debugVar(s2)
        mipx.debugVar(e2)



from math import inf
from typing import Dict, Generator, List, Literal, Optional, Tuple, Union
from ortools.sat.python import cp_model

from .func import get_update_lub
from .utilx import pre_condition
from .interface_ import ITupledict, IVar
from .commonsolver import CommonModelSolver
from .constants import CmpType, ObjType, OptimizationStatus, Params, Vtype

INFINITY = float("inf")
INFINITY_CP = cp_model.INT32_MAX
Real = Union[float, int]


class Model:
    """模型接口"""

    def __init__(
        self,
        solver_id: Literal[
            "SCIP", "CBC", "SAT", "CLP", "CPLEX", "GLOP", "BOP", "GLPK"
        ] = "SCIP",
        name="",
    ):
        super().__init__()
        self.solver_id = solver_id
        if solver_id == "CPLEX":
            try:
                from .cplexmpsolver import CplexMpSolver

                self.__model = CplexMpSolver(name=name)
                if not self.__model.has_cplex_env:
                    raise Exception("No CPLEX environment found")
            except ImportError:
                raise Exception("未找到cplex 库，请安装")
            except:
                raise Exception("未知错误")
        else:
            self.__model = CommonModelSolver(solver_id=solver_id, name=name)
        self._is_cp = solver_id in ["CP", "CPLEX_CP"]
        self._allkpis = {}
        self._multi_object_expr = []

    @property
    def Params(self) -> Params:
        return self.__model.Params

    @property
    def name(self) -> str:
        return self.__model.name

    @property
    def Infinity(self) -> float:
        return self.__model.Infinity

    def sum(self, expr_array):
        return self.Sum(expr_array)

    def Sum(self, expr_array):
        return self.__model.Sum(expr_array)

    def setHint(self, start: Dict[IVar, Real]):
        self.__model.setHint(start)  # type: ignore

    def setTimeLimit(self, time_limit_seconds: int):
        """
        设置程序最大运行时间
        :param time_limit_seconds: 秒
        """
        self.__model.setTimeLimit(time_limit_seconds)

    def wall_time(self) -> float:
        """
        求解所花费的时间
        :return: 求解所花费的时间，单位毫秒
        """
        return self.__model.wall_time()

    @property
    def WallTime(self) -> float:
        return self.wall_time()

    def iterations(self):
        """

        :return: 算法迭代的次数
        """
        return self.__model.iterations()

    def nodes(self) -> Real:
        """

        :return: 节点数
        """
        return self.__model.nodes()

    def addVar(
        self, lb=0, ub=None, vtype: Vtype = Vtype.CONTINUOUS, name: str = ""
    ) -> IVar:
        """
        创建变量
        :param lb: 变量下界
        :param ub: 变量上界
        :param vtype: 变量类型： Vtype.CONTINUOUS（连续）,Vtype.BINARY(0-1变量), Vtype.BINARY（整数变量）
        :param name:变量名
        :return:变量实体
        """
        lb, ub = get_update_lub(self.solver_id, self._is_cp, lb, ub, vtype)
        if vtype == Vtype.INTEGER:
            lb = int(lb + 0.5)
            if ub != inf:
                ub = int(ub)
        return self.__model.addVar(lb=lb, ub=ub, vtype=vtype, name=name)

    def addVars(
        self,
        *indices,
        lb: Real = 0,
        ub=None,
        vtype: Vtype = Vtype.CONTINUOUS,
        name: str = "",
    ) -> ITupledict:
        """
        创建多维变量
        :param indices:多维的参数，如addVars(1,2),addVars(mList,nList),addVars([1,2,3],[3,4,5])等。
        :param lb:变量下界
        :param ub:变量上界
        :param vtype:变量类型： Vtype.CONTINUOUS（连续）,Vtype.BINARY(0-1变量), Vtype.BINARY（整数变量）
        :param name:变量名
        :return:tupledict类型
        """
        pre_condition(len(indices) > 0, "addVars中多维参数缺失")
        lb, ub = get_update_lub(self.solver_id, self._is_cp, lb, ub, vtype)
        return self.__model.addVars(*indices, lb=lb, ub=ub, vtype=vtype, name=name)

    def getVars(self) -> List[IVar]:
        """
        获取所有的变量对象
        :return:
        """
        return self.__model.getVars()

    def getVar(self, i) -> IVar:
        """
        获取第i个变量
        :param i:
        :return:
        """
        return self.__model.getVar(i)

    def addConstr(self, lin_expr, name=""):
        """
        向模型添加约束条件，
        :param lin_expr:线性约束表达式
        :param name: 约束名称
        :return:
        """
        return self.__model.addConstr(lin_expr, name=name)

    def addConstrs(self, lin_exprs: Union[List, Generator, Tuple], name=""):
        """
        向模型添加多个约束条件，
        :param lin_exprs: 线性约束表达式集合 可以为列表或者元组。
        :param name:名称
        :return:
        """
        return self.__model.addConstrs(lin_exprs, name=name)

    def setObjective(self, expr):
        """
        设置模型的单一目标
        :param expr: 目标表达式
        优化方向。ObjType.MINIMIZE（最小值），ObjType.MAXIMIZE(最大值)
        :return:
        """
        self.__model.setObjective(expr)

    def setObjectiveN(
        self, expr, index: int, priority: int = 0, weight: float = 1, name: str = ""
    ):
        """
        多目标优化，优化最小值
        :param expr: 表达式
        :param index: 目标函数对应的序号 (默认 0，1，2，…), 以 index=0 作为目标函数的值, 其余值需要另外设置参数
        :param priority: 分层序列法多目标决策优先级(整数值), 值越大优先级越高【未实现】
        :param weight: 线性加权多目标决策权重(在优先级相同时发挥作用)
        :param name: 名称
        :return:
        """
        self._multi_object_expr.append((expr, weight, name))
        self.__model.setObjectiveN(
            expr, index, priority=priority, weight=weight, name=name
        )

    def addGenConstrAnd(self, resvar, varList: List[IVar], name=""):
        """
        and 运算。
        addGenConstrAnd(y, [x1, x2]) 表示y = and(x1,x2)。 所有变量均为0-1变量
        即 x1 x2 都为1时,y = 1
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        self.__model.addGenConstrAnd(resvar, varList, name=name)

    def addGenConstrOr(self, resvar: IVar, varList: List[IVar], name=""):
        """
        或运算
        addGenConstrOr(y, [x1, x2]) 表示y = or(x1,x2)。 所有变量均为0-1变量
        即 x1,x2 至少有一个为1时,y = 1
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        self.__model.addGenConstrOr(resvar, varList, name=name)

    def addGenConstrXOr(self, resvar: IVar, varList: List[IVar], name=""):
        """
        异或运算
        addGenConstrXOr(y, [x1, x2])。 所有变量均为0-1变量
        x1 x2 y
        1  1  0
        1  0  1
        0  1  1
        0  0  0
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        self.__model.addGenConstrXOr(resvar, varList, name=name)

    def addGenConstrPWL(
        self,
        var_y: IVar,
        var_x: IVar,
        x_range: List[float],
        y_value: List[float],
        name="",
    ):
        """
        设置分段约束
        model.addGenConstrPWL(var, [1, 3, 5], [1, 2, 4])
        :param var_y: f(x)
        :param var_x:指定变量的目标函数是分段线性
        :param x:  定义分段线性变量的点的范围边界(非减序列)，要求x的范围大于0
        :param y:定义分段线性变量的范围所对应的目标函数的值
        :return:
        """
        self.__model.addGenConstrPWL(var_y, var_x, x_range, y_value, name=name)

    def addGenConstrIndicator(
        self,
        binvar: IVar,
        binval: bool,
        lhs: IVar,
        sense: CmpType,
        rhs,
        M,
        name: str = "",
    ):
        """
        若 binvar 为binval ,则 lhs 与 rhs 之间有sense 的关系
        若M不指定，则程序会给与默认。但仍推荐给出M。程序自动给出的可能会存在问题。
        :param binvar: 0-1变量
        :param binval: bool 常量
        :param lhs:  左侧变量
        :param sense: 等号，大于等于，小于等于
        :param rhs: 右侧常量
        :param M: 大M
        :return:
        """
        self.__model.addGenConstrIndicator(
            binvar, binval, lhs, sense, rhs, M=M, name=name
        )

    def addIndicator(self, binvar: IVar, binval: bool, expr, name: str = ""):
        self.__model.addIndicator(binvar, binval, expr, name)

    def addGenConstrAbs(self, resvar, var_abs: IVar, M, name=""):
        """
        绝对值 resvar = |var_abs|
        :param resvar:
        :param var_abs:
        :param name:
        :return:
        """
        self.__model.addGenConstrAbs(resvar, var_abs, name=name, M=M)

    def addConstrMultiply(self, z: IVar, l: Tuple[IVar, IVar], name=""):
        """
        满足 z = x * y
        其中 x 为0,1变量
        :param l:
        :param z: 变量
        :param name:
        :return:
        """
        self.__model.addConstrMultiply(z, l, name=name)

    def addRange(
        self, expr, min_value: Union[float, int], max_value: Union[float, int], name=""
    ):
        """
        添加范围约束
        :param expr: 表达式
        :param min_value: 最小值
        :param max_value: 最大值
        :param name:名称
        :return:
        """
        self.__model.addRange(expr, min_value, max_value, name=name)

    def addConstrOr(
        self,
        constrs: List,
        ok_num: int = 1,
        cmpType: CmpType = CmpType.EQUAL,
        name: str = "",
        M=None,
    ):
        """
        约束的满足情况,满足的次数
        :param constr: 所有的约束
        :param ok_num:  需要满足的个数，具体则根据cmpType
        :param cmpType: CmpType.LESS_EQUAL CmpType.EQUAL,CmpType.GREATER_EQUAL
        :param name: 名称
        :param M: M值，推荐指定M值。
        :return:
        """
        self.__model.addConstrOr(
            constrs, ok_num=ok_num, cmpType=cmpType, name=name, M=M
        )

    def addKpi(self, kpi_arg, name: str):
        if name in self._allkpis:
            raise Exception(f"KPI名称 {name} 重复")
        self._allkpis[name] = kpi_arg

    def reportKpis(self):
        """
        输出所有KPI
        :return:
        """
        print("===========all kpis=============")
        for name, kpi_arg in self._allkpis.items():
            value = self.valueExpression(kpi_arg)
            print(f"{name} = {value}")
        print("================================")

    def kpiByName(self, name: str):
        """根据名称获取KPI"""
        kpi = self._allkpis.get(name)
        if kpi is None:
            raise Exception(f"KPI名称 {name} 不存在")
        return kpi

    def kpiValueByName(self, name: str):
        """根据名称获取KPI的值"""
        kpi = self.kpiByName(name)
        return self.valueExpression(kpi)

    def removeKpi(self, name: str):
        """
        移除KPI
        :param name:
        :return:
        """
        if name not in self._allkpis:
            raise Exception(f"KPI名称 {name} 不存在")
        del self._allkpis[name]

    def clearKpis(self):
        """
        清除所有KPI
        :return:
        """
        self._allkpis.clear()

    @property
    def numOfKpis(self):
        return len(self._allkpis)

    def numVars(self, varType: Optional[Vtype] = None) -> int:
        """
        变量个数
        :return:
        """
        return self.__model.numVars(varType)

    def statistics(self):
        """
        统计模型信息
        :return:
        """
        print("statistics：\n========")
        print(f"numConstraints：{self.numConstraints()}")
        print(f"numVars：{self.numVars()}")
        print(
            f"numVars of int：{self.numVars(Vtype.INTEGER)+self.numVars(Vtype.BINARY)}"
        )
        print("========")

    def reportMultiObjValue(self):
        """
        展示多目标优化的目标值
        """
        print("========MultiObjValue===========")
        for expr, weight, name in self._multi_object_expr:
            value = self.valueExpression(expr)
            print(f"{name} : {round(value*weight, 2)}")
        print("================================")

    def numConstraints(self) -> int:
        """
        约束个数
        :return:
        """
        return self.__model.numConstraints()

    def write(self, filename: str, obfuscated=False):
        """
        写入到文件
        :param filename:文件名，支持后缀 .lp .mps .proto(目前有问题)
        :param obfuscated: 是否混淆，默认不混淆
        :return:
        """
        self.__model.write(filename, obfuscated=obfuscated)

    def read(self, path: str):
        self.__model.read(path)

    @property
    def ObjVal(self) -> Real:
        """目标值"""
        return self.__model.ObjVal

    def optimize(self, obj_type: ObjType = ObjType.MINIMIZE) -> OptimizationStatus:
        """
        优化目标
        :param time_limit_milliseconds:最大运行时长
        :param obj_type:优化目标。ObjType.MINIMIZE（最小值），ObjType.MAXIMIZE(最大值)
        :param enable_output: 是否显示gap日志。
        :return:
        """
        return self.__model.optimize(obj_type=obj_type)

    def clear(self):
        self.clearKpis()
        self._multi_object_expr.clear()
        self.__model.clear()

    def close(self):
        self.__model.close()

    def valueExpression(self, expression):
        """
        计算表达式的值。
        :param expression:
        :return:
        """
        if isinstance(expression, (int, float)):
            return expression
        return self.__model.valueExpression(expression)

    def newIntervalVar(self, start, size, end, name=""):
        """
        创建变量： start+size=end

        Args:
            start (_type_): 开始
            size (_type_): 大小
            end (_type_): 结束
            name (str, optional): 名称. Defaults to "".
        """
        return self.__model.newIntervalVar(start, size, end, name=name)

    def addNoOverlap(self, interval_vars: List, M: int):
        """
        互相之间不重复

        Args:
            interval_vars (List): 间隔变量
        """
        self.__model.addNoOverlap(interval_vars, M)

    def setNumThreads(self, num_theads: int):
        """
        设置线程的个数

        Args:
            num_theads (int): 线程个数
        """
        self.__model.setNumThreads(num_theads)



# -*- coding: utf-8 -*-
# @Time    : 2023/4/12 9:12
# @Author  : luyi
"""
约束规划
"""

from typing import Optional, Union, List, Dict, Tuple

from docplex.cp.model import true
from ortools.sat.python import cp_model
from ortools.util.python.sorted_interval_list import Domain

from .constants import Vtype, ObjType, OptimizationStatus, CmpType
from .interface_ import IVar, ICpModel
from .tupledict import tupledict
from .utilx import (
    is_list_or_tuple,
    get_combinations,
    name_str,
    is_generator,
    check_bool_var,
    is_bool_var,
)
from .variable import CpVar

INFINITY = cp_model.INT32_MAX

Real = Union[float, int]


class CpModelSolver(ICpModel):

    def __init__(self, name=""):
        super().__init__("CP")
        self.name = name
        self.__model = cp_model.CpModel()
        self.__solver: cp_model.CpSolver = cp_model.CpSolver()
        self.__line_expr_object: Optional[cp_model.LinearExpr] = None
        self.__all_vars = []
        self.__all_constr = []
        self._flag_objective_n = False
        self._flag_objective = False

    @property
    def Infinity(self) -> int:
        return INFINITY

    @staticmethod
    def Sum(expr_array):
        return cp_model.LinearExpr.Sum(expr_array)

    def setHint(self, start: Dict[IVar, Real]):
        for var, num in start.items():
            self.__model.AddHint(var, num)  # type: ignore

    def setTimeLimit(self, time_limit_seconds: int):
        self.__solver.parameters.max_time_in_seconds = int(time_limit_seconds)

    def wall_time(self) -> int:
        return int(self.__solver.WallTime()) * 1000

    def iterations(self) -> Real:
        return 0

    def nodes(self) -> Real:
        return self.__solver.NumBranches()

    def addVar(
        self,
        lb: int = 0,
        ub: int = INFINITY,
        vtype: Vtype = Vtype.BINARY,
        name: str = "",
    ) -> IVar:
        """
            Add a decision variable to a model.
        :param lb: Lower bound for new variable.
        :param ub: Upper bound for new variable.
        :param vtype: variable type for new variable(Vtype.CONTINUOUS, Vtype.BINARY, Vtype.INTEGER).
        :param name: Name for new variable.
        :return: variable.
        """
        tempModel = self.__model._CpModel__model  # type: ignore
        var1: Optional[CpVar] = None
        if vtype == Vtype.CONTINUOUS or vtype == Vtype.INTEGER:
            var1 = CpVar(tempModel, Domain(lb, ub), False, name, self)
        elif vtype == Vtype.BINARY:
            var1 = CpVar(tempModel, Domain(lb, ub), True, name, self)
        self.__all_vars.append(var1)
        return var1

    def addVars(
        self,
        *indices,
        lb: int = 0,
        ub: int = INFINITY,
        vtype: Vtype = Vtype.INTEGER,
        name: str = "",
    ) -> tupledict:
        li = []
        for ind in indices:
            if isinstance(ind, int):
                ind = [i for i in range(ind)]
            elif is_list_or_tuple(ind):
                pass
            elif isinstance(ind, str):
                ind = [ind]
            else:
                raise ValueError("error input")
            li.append(ind)
        all_keys_tuple = get_combinations(li)
        tu_dict = tupledict(
            [
                [key, self.addVar(lb, ub, vtype, name_str(name, key))]
                for key in all_keys_tuple
            ]
        )
        return tu_dict

    def getVars(self) -> List[IVar]:
        return self.__all_vars

    def getVar(self, i) -> IVar:
        return self.__all_vars[i]

    def addConstr(self, lin_expr, name=""):
        self.__all_constr.append(lin_expr)
        self.__model.Add(lin_expr)

    def addConstrs(self, lin_exprs, name=""):
        # 检查下是否可以迭代。
        if not is_list_or_tuple(lin_exprs) and not is_generator(lin_exprs):
            raise RuntimeError("constraint conditions are not a set or list")
        for i, lin_expr in enumerate(lin_exprs):
            self.addConstr(lin_expr)

    def setObjective(self, expr):
        self._flag_objective = True
        if self._flag_objective_n:
            raise RuntimeError(
                "setObjective and setObjectiveN can only be used for one of them"
            )
        self.__model.Minimize(expr)
        self.__line_expr_object = expr

    def setObjectiveN(
        self, expr, index: int, priority: int = 0, weight: float = 1, name: str = ""
    ):
        """

        :param expr:
        :param index:
        :param priority:
        :param weight:
        :param name:
        :return:
        """
        self._flag_objective_n = True
        if self._flag_objective:
            raise RuntimeError(
                "setObjective and setObjectiveN can only be used for one of them"
            )
        if self.__line_expr_object is None:
            self.__line_expr_object = expr * weight
            return
        self.__line_expr_object = self.__line_expr_object + expr * weight

    def addGenConstrAnd(self, resvar, varList: List[IVar], name=""):
        """
        和 addGenConstrAnd(y, [x1,x2])
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        check_bool_var(resvar, varList)
        for var in varList:
            self.addConstr(resvar <= var)
        self.addConstr(resvar >= self.Sum(varList) -
                       len(varList) + 1)  # type: ignore

    def addGenConstrOr(self, resvar: IVar, varList: List[IVar], name=""):
        check_bool_var(resvar, varList)
        for var in varList:
            self.addConstr(resvar >= var)
        self.addConstr(resvar <= self.Sum(varList))

    def addGenConstrMin(self, target: IVar, varList: List[IVar]):
        self.__model.add_min_equality(target, varList)  # type: ignore

    def addGenConstrMax(self, target: IVar, varList: List[IVar]):
        """
        最小值约束
        """
        self.__model.add_max_equality(target, varList)  # type: ignore

    def addGenConstrDivision(self, target, num, denom):
        """Adds `target == num // denom` (integer division rounded towards 0)."""
        self.__model.add_division_equality(target, num, denom)

    def addGenConstrModulo(self, target, num, denom):
        """Adds `target = expr % mod`."""
        self.__model.add_modulo_equality(target, num, denom)

    def addGenConstrMultiplication(
        self,
        target,
        *expressions,
    ):
        """Adds `target == expressions[0] * .. * expressions[n]`."""
        self.__model.add_multiplication_equality(target, *expressions)

    def addGenConstrXOr(self, resvar: IVar, varList: List[IVar], name=""):

        if len(varList) != 2:
            raise ValueError("length of vars must be 2")
        check_bool_var(resvar, varList)
        self.addConstr(resvar >= varList[0] - varList[1])
        self.addConstr(resvar >= varList[1] - varList[0])
        self.addConstr(resvar <= varList[0] + varList[1])
        self.addConstr(resvar <= 2 - varList[0] - varList[1])

    def addGenConstrPWL(
        self,
        var_y: IVar,
        var_x: IVar,
        x_range: List[int],
        y_value: List[int],
        name="",
    ):
        raise NotImplementedError()

    def addGenConstrIndicator(
        self,
        binvar: IVar,
        binval: bool,
        lhs: IVar,
        sense: CmpType,
        rhs: int,
        M,
        name: str = "",
    ):
        # var = binvar if binval is True else binvar.Not()
        # if sense == CmpType.GREATER_EQUAL:  # 大于等于
        #     self.__model.Add(lhs - rhs >= 0).OnlyEnforceIf(var)
        # elif sense == CmpType.EQUAL:
        #     self.__model.Add(lhs - rhs == 0).OnlyEnforceIf(var)
        # else:
        #     self.__model.Add(lhs - rhs <= 0).OnlyEnforceIf(var)
        #  根据不完全的测试，下面的代码优于上面的代码
        if M is None:
            M = abs(rhs) + 100  # TODO:待优化
        if binval is True:
            z = 1 - binvar
        else:
            z = binvar
        if sense == CmpType.GREATER_EQUAL:
            self.addConstr(lhs + M * z >= rhs)
        elif sense == CmpType.EQUAL:
            self.addConstr(lhs + M * z >= rhs)
            self.addConstr(lhs - M * z <= rhs)
        else:
            self.addConstr(lhs - M * z <= rhs)

    def addIndicator(self, binvar: IVar, binval: bool, expr, name: str = ""):
        raise RuntimeError("not implemented")

    def addGenConstrAbs(self, resvar, var_abs: IVar):
        self.__model.AddAbsEquality(resvar, var_abs)  # type: ignore

    def addConstrMultiply(self, z: IVar, l: Tuple[IVar, IVar], name=""):
        """x * y = z"""
        super().addConstrMultiply(z, l)
        x = l[0]
        y = l[1]
        if not is_bool_var(x) and not is_bool_var(y):
            raise RuntimeError("At least one binary variable is required.")
        if is_bool_var(y):
            x, y = y, x
        M = y.Ub
        self.addConstr(z <= y)
        self.addConstr(z <= x * M)
        self.addConstr(z >= y + (x - 1) * M)

    def addRange(
        self, expr, min_value: Union[float, int], max_value: Union[float, int], name=""
    ):
        self.__model.AddLinearConstraint(
            expr, min_value, max_value)  # type: ignore

    def addConstrOr(
        self,
        constrs: List[cp_model.BoundedLinearExpression],
        ok_num: int = 1,
        cmpType: CmpType = CmpType.EQUAL,
        name: str = "",
        M=None,
    ):
        """
        约束的满足情况

        :param constr: 所有的约束
        :param ok_num: 需要满足的个数，具体则根据cmpType
        :param cmpType: CmpType.LESS_EQUAL CmpType.EQUAL,CmpType.GREATER_EQUAL
        :param name: 名称
        :return:
        """
        constr_num = len(constrs)
        x = []
        for i in range(constr_num):
            x.append(self.addVar(vtype=Vtype.BINARY))
            constr = constrs[i]
            try:
                lb, ub = constr.Bounds()  # type: ignore
                expr = constr.Expression()  # type: ignore
            except:
                lb, ub = constr.bounds()
                expr = constr.expression()
            tempM = 100000000
            if M is not None:
                tempM = M
            if lb > -INFINITY:  # 若大于
                self.addConstr(expr + tempM * (1 - x[i]) >= lb)
            if ub < INFINITY:

                self.addConstr(expr - tempM * (1 - x[i]) <= ub)
        if cmpType == CmpType.EQUAL:
            self.addConstr(self.Sum(x) == ok_num)
        elif cmpType == CmpType.GREATER_EQUAL:
            self.addConstr(self.Sum(x) >= ok_num)  # type: ignore
        elif cmpType == CmpType.LESS_EQUAL:
            self.addConstr(self.Sum(x) <= ok_num)  # type: ignore
        else:
            raise Exception("error value of cmpType")

    def numVars(self, vtype: Optional[Vtype] = None) -> int:
        if vtype is Vtype.CONTINUOUS:
            return 0
        return len(self.__all_vars)

    def numConstraints(self) -> int:
        return len(self.__all_constr)

    def write(self, filename: str, obfuscated=False):
        self.__model.ExportToFile(filename)

    def read(self, path: str):
        raise RuntimeError("not implemented")

    @property
    def ObjVal(self) -> Real:
        return self.__solver.ObjectiveValue()

    def optimize(self, obj_type: ObjType = ObjType.MINIMIZE) -> OptimizationStatus:
        solver = self.__solver
        self._set_objective(obj_type)
        self._set_params()
        status = solver.Solve(self.__model)
        if status == cp_model.OPTIMAL:
            result = OptimizationStatus.OPTIMAL
        elif status == cp_model.INFEASIBLE:
            result = OptimizationStatus.INFEASIBLE
        elif status == cp_model.UNKNOWN:
            result = OptimizationStatus.UNBOUNDED
        elif status == cp_model.FEASIBLE:
            result = OptimizationStatus.FEASIBLE
        else:
            result = OptimizationStatus.ERROR
        return result

    def clear(self):
        self.__line_expr_object = None
        self.__all_vars.clear()
        self.__all_constr.clear()
        self._flag_objective = False
        self._flag_objective_n = False
        self.__model.ClearObjective()
        self.__model.ClearHints()
        self.__model.ClearAssumptions()

    def close(self):
        self.clear()
        self.__solver = None  # type: ignore

    def _set_objective(self, obj_type: ObjType):
        if self.__line_expr_object is not None:
            if obj_type == ObjType.MINIMIZE:
                self.__model.Minimize(self.__line_expr_object)
            else:
                self.__model.Maximize(self.__line_expr_object)

    def _set_params(self):
        if self.Params.TimeLimit:
            self.setTimeLimit(self.Params.TimeLimit)
        if self.Params.EnableOutput:
            self.__solver.parameters.log_search_progress = True
        if self.Params.MIPGap:
            self.__solver.parameters.absolute_gap_limit = self.Params.MIPGap

    # 独有。其他方法待添加

    def addNoOverlap(self, interval_vars, M):
        """
        相邻不重复。

        :param interval_vars:
        :return:
        """
        self.__model.AddNoOverlap(interval_vars)

    def newIntervalVar(self, start, size, end, name=""):
        if type(start) == type(0.1):
            start = int(start)
        if type(end) == type(0.1):
            end = int(end)
        if type(size) == type(0.1):
            size = int(size)
        return self.__model.NewIntervalVar(start, size, end, name)

    def valueExpression(self, expression):
        return self.__solver.Value(expression)

    def setNumThreads(self, num_theads: int):
        self.__solver.parameters.num_workers = num_theads

    def addElement(self, index, variables, target):
        self.__model.add_element(index, variables, target)

    def addCircuit(self, arcs):
        self.__model.add_circuit(arcs)

    def addAllowedAssignments(self, variables, tuples_list):
        self.__model.add_allowed_assignments(variables, tuples_list)

    def addForbiddenAssignments(self, variables, tuples_list):
        self.__model.add_forbidden_assignments(variables, tuples_list)

    def addInverse(self, variables, inverse_variables):
        self.__model.add_inverse(variables, inverse_variables)

    def addMapDomain(self, var, bool_var_array, offset):
        self.__model.add_map_domain(var, bool_var_array, offset)

    def addImplication(self, a, b):
        self.__model.add_implication(a, b)

    def addBoolTrueOr(self, literals):
        self.__model.add_bool_or(literals)

    def addAtLeastOneIsTrue(self, literals):
        self.__model.add_at_least_one(literals)

    def addAtMostOneIsTrue(self, literals):
        self.__model.add_at_most_one(literals)

    def addExactlyNumIsTrue(self, literals, num: int = 1):
        if num == 1:
            self.__model.AddExactlyOne(literals)
        else:
            self.addConstr(self.Sum(literals) == num)

    def addBoolTrueAnd(self, literals):
        self.__model.add_bool_and(literals)



# -*- coding: utf-8 -*-
# @Time    : 2023/4/1 10:39
# @Author  : luyi
from functools import reduce
from typing import Dict, List
from .utilx import tuple_fuzz_match_list, is_list_or_tuple, get_length
from .tuplelist import tuplelist
from .interface_ import ITupledict
from warnings import warn


class tupledict(ITupledict):
    """
    tupledict is a subclass of dict where
      the keys are a tuplelist.
    """

    def __init__(self, *args, **kwargs):
        self.dim = None
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        # if is_list_or_tuple(value):  # 不复杂化，，仅支持为一个参数。一般为变量var
        #     raise ValueError("tupledict的value值应为单一对象")
        if self.dim is None:
            self.dim = get_length(key)
        super().__setitem__(key, value)

    def prod(self, coeff: Dict, *pattern):  # type:ignore
        """
        coeff为一个dict类型，指定待计算的元素的系数。coeff的key要与待计算的集合中的key能对应
        :param coeff:
        :param pattern:
        :return:
        """
        warn(f"call to deprecated function,use quickprod function")
        tList = tuplelist(self.keys())
        tl = tList.select(*pattern)
        if len(tl) == 0:
            return 0
        pl = []
        for key in tl:
            ok, k = tuple_fuzz_match_list(key, coeff.keys())
            if ok:
                pl.append(self[key] * coeff[k])
        return reduce(lambda a, b: a + b, pl)

    def select(self, *pattern):
        warn(f"call to deprecated function,use quickselect function")
        tlist = tuplelist(self.keys())
        keys = tlist.select(*pattern)
        return [self[key] for key in keys]

    def sum(self, *pattern):  # type:ignore
        warn(f"call to deprecated function,use quicksum function")
        v = self.select(*pattern)
        if len(v) == 0:
            return 0
        return reduce(lambda a, b: a + b, v)

    def quickprod(self, coeff: Dict, *pattern):  # type:ignore
        tList = tuplelist(self.keys())
        tl = tList.quickselect(*pattern)
        if len(tl) == 0:
            return 0
        pl = []
        for key in tl:
            ok, k = tuple_fuzz_match_list(key, coeff.keys())
            if ok:
                pl.append(self[key] * coeff[k])
        item = self[tl[0]]
        return item._solver.Sum(pl)

    def quickselect(self, *pattern) -> List:
        """
        快速挑选符合模式的items
        """
        dim = len(pattern)
        if dim == 0:
            return list(self.data.values())  # type: ignore
        li = []
        if self.dim == 1:
            for key in self.data.keys():
                p = pattern[0]
                if p == "*" or p == key:
                    li.append(self[key])
            return li
        for key in self.data.keys():
            ok = True
            for i, k in enumerate(key):
                p = pattern[i]
                if p == "*":
                    continue
                if p != k:
                    ok = False
                    break
            if ok:
                li.append(self[key])
        return li

    def quicksum(self, *pattern):  # type:ignore
        v = self.quickselect(*pattern)
        if len(v) == 0:
            return 0
        item = v[0]
        return item._solver.Sum(v)


def multidict(data: Dict):
    """

    :param data:
    :return:
    """
    """
    ROUTINE:
        multidict(data)
      PURPOSE:
        Split a single dictionary into multiple dictionaries.
      ARGUMENTS:
        data: A dictionary that maps each key to a list of 'n' values.
      RETURN VALUE:
        A list of the shared keys, followed by individual tupledicts.
      EXAMPLE:
        keys, [dict1, dict2] = multidict( {
                 'key1': [1, 2],
                 'key2': [1, 3],
                 'key3': [1, 4] } )
    """
    t_list = tuplelist(data.keys())
    num = None
    for key, items in data.items():
        _num = get_length(items)
        if num is None:
            num = _num
        else:
            if num != _num:
                raise ValueError("length of values should be same")
    if num is None:
        raise ValueError("multidict error")
    t_dicts = [
        tupledict(
            {
                key: data[key][i] if is_list_or_tuple(data[key]) else data[key]
                for key in t_list
            }
        )
        for i in range(num)
    ]
    if num == 1:
        return t_list, t_dicts[0]
    return t_list, *t_dicts



# -*- coding: utf-8 -*-
# @Time    : 2023/4/12 9:12
# @Author  : luyi
"""
约束规划
"""

from typing import Optional, Sequence, Union, List, Dict, Tuple

from docplex.cp.config import CpoParameters
from docplex.cp.model import CpoModel, Type_Int
from docplex.cp.expression import (
    CpoExpr,
    CpoFunctionCall,
    Type_Float,
)
from .variable_cplex import CPlexCpoVar
from .constants import Vtype, ObjType, OptimizationStatus, CmpType
from .interface_ import IVar, ICpModel
from .tupledict import tupledict
from .utilx import (
    is_list_or_tuple,
    get_combinations,
    name_str,
    check_bool_var,
    is_bool_var,
)


Real = Union[float, int]
INFINITY = 1e10


class CplexCpSolver(ICpModel):
    """
    约束规划CPLEX 未实现

    :param _type_ IModel: _description_
    """

    def __init__(self, name=""):
        super().__init__("CPLEX_CP")
        self.name = name
        self.__model = CpoModel(self.name)
        self._solver = None
        self.__line_expr_object = None
        self._flag_objective_n = False
        self._flag_objective = False

    def _get_execfile_path(self) -> str:
        import os

        cplex_studio_dir = os.getenv("CPLEX_STUDIO_CPOPTIMIZER_DIR")
        if cplex_studio_dir:
            from pathlib import Path

            cpoptimizer_path = Path(cplex_studio_dir)
            if cpoptimizer_path.exists():
                return str(cplex_studio_dir)
            else:
                raise RuntimeError(
                    f"环境变量[CPLEX_STUDIO_CPOPTIMIZER_DIR={cplex_studio_dir}]指定的目录不存在"
                )
        else:
            raise RuntimeError(
                "请设置环境变量[CPLEX_STUDIO_CPOPTIMIZER_DIR]或者设置model.Params.CpoptimizerPath"
            )

    @property
    def Infinity(self) -> float:
        return INFINITY

    def Sum(self, expr_array) -> "IVar":  # type: ignore
        return self.__model.sum(expr_array)  # type: ignore

    def setHint(self, start: Dict[IVar, Real]):
        raise RuntimeError("not implemented")

    def setTimeLimit(self, time_limit_seconds: int):
        self.Params.TimeLimit = time_limit_seconds

    def wall_time(self) -> int:
        solver_details: SolveDetails = self.__model.solve_details  # type: ignore
        return solver_details.time * 1000

    def iterations(self) -> Real:
        solver_details: SolveDetails = self.__model.solve_details  # type: ignore
        return solver_details.nb_iterations

    def nodes(self) -> Real:
        solver_details: SolveDetails = self.__model.solve_details  # type: ignore
        return solver_details.nb_nodes_processed

    def addVar(
        self, lb=None, ub=None, vtype: Vtype = Vtype.INTEGER, name: str = ""
    ) -> IVar:
        """
            Add a decision variable to a model.
        :param lb: Lower bound for new variable.
        :param ub: Upper bound for new variable.
        :param vtype: variable type for new variable(Vtype.CONTINUOUS, Vtype.BINARY, Vtype.INTEGER).
        :param name: Name for new variable.
        :return: variable.
        """
        if lb is None:
            lb = 0
        if vtype == Vtype.CONTINUOUS or vtype == Vtype.INTEGER:
            var1 = CPlexCpoVar(True, self, lb, ub, name)
        else:
            var1 = CPlexCpoVar(False, self, lb, ub, name)
        return var1

    def addVars(
        self,
        *indices,
        lb=None,
        ub=None,
        vtype: Vtype = Vtype.INTEGER,
        name: str = "",
    ) -> tupledict:

        li = []
        for ind in indices:
            if isinstance(ind, int):
                ind = [i for i in range(ind)]
            elif is_list_or_tuple(ind):
                pass
            elif isinstance(ind, str):
                ind = [ind]
            else:
                raise ValueError("error input")
            li.append(ind)
        all_keys_tuple = get_combinations(li)
        tu_dict = tupledict(
            [
                [key, self.addVar(lb, ub, vtype, name_str(name, key))]
                for key in all_keys_tuple
            ]
        )
        return tu_dict

    def getVars(self) -> List[IVar]:
        return self.__model.get_all_variables()  # type: ignore

    def getVar(self, i) -> IVar:
        raise RuntimeError("not implemented")

    def addConstr(self, lin_expr, name=""):
        self.__model.add_constraint(lin_expr)

    def addConstrs(self, lin_exprs, name=""):
        # 检查下是否可以迭代。
        self.__model.add(lin_exprs)

    def setObjective(self, expr):
        self._flag_objective = True
        if self._flag_objective_n:
            raise RuntimeError(
                "setObjective and setObjectiveN can only be used for one of them"
            )
        self.__line_expr_object = expr

    def setObjectiveN(
        self, expr, index: int, priority=None, weight: float = 1, name: str = ""
    ):
        """
        :param expr:
        :param index:
        :param priority:
        :param weight:
        :param name:
        :return:
        """
        self._flag_objective_n = True
        if self._flag_objective:
            raise RuntimeError(
                "setObjective and setObjectiveN can only be used for one of them"
            )
        if self.__line_expr_object is None:
            self.__line_expr_object = expr * weight
            return
        self.__line_expr_object = self.__line_expr_object + expr * weight

    def addGenConstrAnd(self, resvar, varList: List[IVar], name=""):
        """
        和 addGenConstrAnd(y, [x1,x2])
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        self.addConstr(resvar == self.__model.logical_and())  # type: ignore

    def addGenConstrOr(self, resvar: IVar, varList: List[IVar], name=""):
        self.addConstr(resvar == self.__model.logical_or(
            *varList))  # type: ignore

    def addGenConstrXOr(self, resvar: IVar, varList: List[IVar], name=""):
        if len(varList) != 2:
            raise ValueError("length of vars must be 2")
        check_bool_var(resvar, varList)
        self.addConstr(resvar >= varList[0] - varList[1])
        self.addConstr(resvar >= varList[1] - varList[0])
        self.addConstr(resvar <= varList[0] + varList[1])
        self.addConstr(resvar <= 2 - varList[0] - varList[1])

    def addGenConstrPWL(
        self, var_y: IVar, var_x: IVar, x_range: List[int], y_value: List[int], name=""
    ):
        raise NotImplementedError()

    def addGenConstrIndicator(
        self,
        binvar: IVar,
        binval: bool,
        lhs: IVar,
        sense: CmpType,
        rhs: int,
        M,
        name: str = "",
    ):
        # TODO:
        raise RuntimeError("not implemented")

    def addIndicator(self, binvar: IVar, binval: bool, expr, name=None):
        raise RuntimeError("not implemented")

    def addGenConstrAbs(self, resvar, var_abs: IVar):  # type: ignore
        self.addConstr(self.__model.abs(var_abs) == resvar)  # type: ignore

    def addConstrMultiply(self, z: IVar, l: Tuple[IVar, IVar], name=""):
        """x * y = z"""
        super().addConstrMultiply(z, l)
        x = l[0]
        y = l[1]
        if not is_bool_var(x) and not is_bool_var(y):
            raise RuntimeError("At least one binary variable is required.")
        if is_bool_var(y):
            x, y = y, x
        M = y.Ub
        self.addConstr(z <= y)
        self.addConstr(z <= x * M)
        self.addConstr(z >= y + (x - 1) * M)

    def addRange(
        self, expr, min_value: Union[float, int], max_value: Union[float, int], name=""
    ):
        self.addConstr(expr >= min_value)
        self.addConstr(expr <= max_value)

    def addConstrOr(
        self,
        constrs: List,
        ok_num: int = 1,
        cmpType: CmpType = CmpType.EQUAL,
        name: str = "",
        M=None,
    ):
        """
         约束的满足情况

        :param constr: 所有的约束
        :param ok_num: 需要满足的个数，具体则根据cmpType
        :param cmpType: CmpType.LESS_EQUAL CmpType.EQUAL,CmpType.GREATER_EQUAL
        :param name: 名称
        :return:
        """
        constr_num = len(constrs)
        x = []
        for i in range(constr_num):
            x.append(self.addVar(vtype=Vtype.BINARY))
            constr = constrs[i]
            lb, ub = constr.Bounds()
            expr = constr.Expression()
            tempM = 100000000
            if M is not None:
                tempM = M
            if lb > -INFINITY:  # 若大于
                self.addConstr(expr + tempM * (1 - x[i]) >= lb)
            if ub < INFINITY:
                self.addConstr(expr - tempM * (1 - x[i]) <= ub)
        if cmpType == CmpType.EQUAL:
            self.addConstr(self.Sum(x) == ok_num)
        elif cmpType == CmpType.GREATER_EQUAL:
            self.addConstr(self.Sum(x) >= ok_num)
        elif cmpType == CmpType.LESS_EQUAL:
            self.addConstr(self.Sum(x) <= ok_num)  # type: ignore
        else:
            raise Exception("error value of cmpType")

    def numVars(self, vtype: Optional[Vtype] = None) -> int:
        return len(self.__model.get_all_variables())

    def numConstraints(self) -> int:
        return len(self.__model.expr_list)

    def write(self, filename: str, obfuscated=False):
        raise NotImplementedError()

    def read(self, path: str):  # type: ignore
        raise NotImplementedError()

    @property
    def ObjVal(self) -> Real:
        res = self._solver.get_objective_values()  # type: ignore
        if res:
            return res[0]  # type: ignore
        return 0

    def optimize(self, obj_type: ObjType = ObjType.MINIMIZE) -> OptimizationStatus:
        params = CpoParameters()
        if self.Params.TimeLimit:
            params.TimeLimit = self.Params.TimeLimit
        self.__model.set_parameters(params)
        if self.__line_expr_object is not None:
            if obj_type == ObjType.MINIMIZE:
                self.__model.minimize(self.__line_expr_object)
            else:
                self.__model.maximize(self.__line_expr_object)
        trace_log = True if self.Params.EnableOutput else False
        if self.Params.CpoptimizerPath is None:
            cplex_file = self._get_execfile_path()
            self.Params.CpoptimizerPath = cplex_file
        sol = self.__model.solve(
            execfile=self.Params.CpoptimizerPath, trace_log=trace_log
        )
        if sol:  # type: ignore
            self._solver = sol
            return OptimizationStatus.OPTIMAL
        else:
            return OptimizationStatus.ERROR

    def clear(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    # 独有。

    def addNoOverlap(self, interval_vars, M):
        """
        相邻不重复。

        :param interval_vars:
        :return:
        """
        raise RuntimeError("not implemented")

    def newIntervalVar(self, start, size, end, name=""):
        raise RuntimeError("not implemented")

    def valueExpression(self, expression: CpoExpr) -> Real:  # type: ignore
        assert self._solver is not None, "执行optimize后才能进行值评价"
        if expression.is_type(Type_Int) or expression.is_type(Type_Float):
            return expression.value  # type: ignore
        if expression.is_variable():
            try:
                return self._solver[expression]  # type:ignore
            except KeyError:
                print(f"warning:变量{expression}未使用")
            # 此时应该是这个变量没有使用,直接去最小值
            return expression.lb  # type:ignore
        # 此时是表达式
        assert isinstance(
            expression, CpoFunctionCall
        ), f"参数`expression`必须是CpoFunctionCall,传入的是:{expression}"
        assert (
            len(expression.children) == 2
        ), f"CpoFunctionCall下的`expression.children`的长度必须等于2"
        left = expression.children[0]
        right = expression.children[1]
        op = expression.operation
        left_value = self.valueExpression(left)
        right_value = self.valueExpression(right)
        value = 0
        if op.get_cpo_name() == "plus":
            value = left_value + right_value
        elif op.get_cpo_name() == "times":
            value = left_value * right_value
        elif op.get_cpo_name() == "minus":
            value = left_value - right_value
        else:
            raise ValueError(f"暂不支持的操作符{op.get_cpo_name()}")
        return value

    def setNumThreads(self, num_theads: int):
        raise RuntimeError("not implemented")

    def addElement(
        self,
        index: Union[IVar, int],
        variables: Sequence[Union[IVar, int]],
        target: Union[IVar, int],
    ): ...

    def addCircuit(self, arcs): ...

    def addAllowedAssignments(self, variables, tuples_list): ...

    def addForbiddenAssignments(self, variables, tuples_list): ...

    def addInverse(self, variables, inverse_variables): ...

    def addMapDomain(self, var, bool_var_array, offset): ...

    def addImplication(self, a, b): ...

    def addBoolTrueOr(self, literals): ...

    def addAtLeastOneIsTrue(self, literals): ...

    def addAtMostOneIsTrue(self, literals): ...

    def addExactlyNumIsTrue(self, literals, num: int = 1): ...

    def addBoolTrueAnd(self, literals): ...

    def addGenConstrMax(self, target: IVar, varList: List[IVar]):
        """
        最大值约束
        """
        ...

    def addGenConstrMin(self, target: IVar, varList: List[IVar]):
        """
        最小值约束
        """
        ...

    def addGenConstrDivision(self, target, num, denom):
        """Adds `target == num // denom` (integer division rounded towards 0)."""
        ...

    def addGenConstrModulo(self, target, num, denom):
        """Adds `target = expr % mod`."""
        ...

    def addGenConstrMultiplication(
        self,
        target,
        *expressions,
    ): ...



import collections
import platform
from typing import List, Tuple
from itertools import combinations

import numpy as np
import mipx
from matplotlib import pyplot as plt

Line = collections.namedtuple('Point', 'x1 y1 x2 y2')


def draw(lines: List[Line]):
    if platform.system() == "Windows":
        plt.rcParams['font.sans-serif'] = ['SimHei']
    else:
        plt.rcParams['font.sans-serif'] = ['Heiti TC']  # 防止中文乱码
    plt.rcParams['axes.unicode_minus'] = False
    for line in lines:
        xs = [line.x1, line.x2]
        ys = [line.y1, line.y2]
        plt.plot(xs, ys, marker='o', markersize=5)
        plt.plot(xs, ys, linestyle='solid')
    plt.show()


# points = [(1, 3), (2, 4), (5, 2), (8, 12), (6, 4),
#           (3, 8), (4, 10), (4, 5), (5, 8), (8, 6), (1, 7)]


def generate_random_points(n_points, seed=None, x_range=(0, 100), y_range=(0, 100)):
    """
    生成随机的二维坐标点
    - n_points: 生成的点位数量
    - seed: 随机种子（可选，用于复现结果）
    - x_range: X坐标范围（默认0-100）
    - y_range: Y坐标范围（默认0-100）
    """
    if seed is not None:
        np.random.seed(seed)

    x = np.random.uniform(x_range[0], x_range[1], n_points)
    y = np.random.uniform(y_range[0], y_range[1], n_points)

    return list(zip(x, y))


# 示例：生成10个随机点（种子44保证可复现）
points = generate_random_points(
    n_points=13, seed=45,  x_range=(0, 10), y_range=(0, 15))


def calc_distance_matrix():
    # 旅行地坐标

    return {(i1, i2): (points[i1][0]-points[i2][0])**2 + (points[i1][1]-points[i2][1])**2 for i1 in range(len(points)) for i2 in range(len(points))}


def tsp():
    # 创建模型
    model = mipx.Model(solver_id="SCIP", name="TSP")
    model.Params.MIPGap = 0.01
    num = len(points)
    # model = mipx.CpModel(solver_id="CP", name="TSP")
    # 决策变量: x[i][j] 表示是否从 i 地点到 j 地点, 1 表示选择，0 表示不选择
    # x = model.addVars(len(points), len(points), vtype=mipx.BINARY, name="x")
    x = mipx.tupledict({(i1, i2): model.addVar(vtype=mipx.BINARY, name=mipx.name_str(
        'x', i1, i2)) for i1 in range(num) for i2 in range(num) if i1 != i2})
    # x = mipx.tupledict()
    # for i1 in range(num):
    #     for i2 in range(num):
    #         if i1 != i2:
    #             x[i1, i2] = model.addVar(vtype=mipx.BINARY, name=mipx.name_str(
    #                 'x', i1, i2))
    # 约束
    # 1. 每个城市只有一个入边
    # for i in range(len(points)):
    #     model.addConstr(x.quicksum(i, "*") == 1)
    model.addConstrs(x.quicksum(i, "*") == 1
                     for i in range(num))
    # 2. 每个城市只有一个出边
    for j in range(num):
        model.addConstr(x.quicksum("*", j) == 1)
        model.addKpi(x.quicksum("*", j), name=f"KPI_{j}")
    # 3. 防止生成子回路
    for n in range(2, num-1):
        coms = combinations(range(num), n)
        for com in coms:
            li = [x[i1, i2] for i1 in com for i2 in com if i1 != i2]
            model.addConstr(model.sum(li) <= n - 1)
    # 添加目标
    # 最小化总距离
    dist_matrix = calc_distance_matrix()
    model.setObjective(x.quickprod(dist_matrix, "*", "*"))
    model.statistics()
    status = model.optimize()
    if status == mipx.OptimizationStatus.OPTIMAL:
        print("Optimal solution found")
        # mipx.debugVar(x)
        lines = []
        model.reportKpis()
        for (i1, i2), x_var in x.items():
            if x_var.X == 1:
                p1 = points[i1]
                p2 = points[i2]
                lines.append(Line(p1[0], p1[1], p2[0], p2[1]))
        draw(lines)


if __name__ == '__main__':
    tsp()



import mipx


def test_model():
    # 创建模型
    # 支持 "SCIP", "CBC", "SAT", "CLP", "CPLEX","GUROBI","CP",....
    model = mipx.Model(solver_id="SAT", name="test_model")
    #! 参数配置
    model.Params.TimeLimit = 1000  # 设置时间限制
    model.Params.MIPGap = 0.01  # 设置MIPGap
    model.Params.EnableOutput = False  # 开启输出
    model.Params.Precision = 1e-6  # 设置精度
    # cpmodel = mipx.CpModel(solver_id="CP", name="test_cpmodel")
    #  添加变量
    # *********************************
    #! 单个变量
    x1 = model.addVar(name="x")  # 单个变量
    # CONTINUOUS（连续）,BINARY(0-1变量),BINARY（整数变量）
    x2 = model.addVar(lb=0, ub=10, vtype=mipx.BINARY, name="x1")
    #! 多维变量
    x3 = model.addVars(2, vtype=mipx.INTEGER, name="x2")  # x3[0],x3[1]
    x4 = model.addVars(2, 3, lb=1, ub=3, name="x4")  # 创建了2行3列的变量
    x5 = model.addVars([1, 2, 3], 3)  # 创建了3h行3列的变量, x5[1,0] x5[1,1] x5[1,2]...
    x6 = model.addVars(2, 2, 3)  # 三维变量
    #! 灵活创建变量
    x = mipx.tupledict({i: model.addVar() for i in range(10)})
    x = mipx.tupledict()  # 也可以使用 x = {}创建,但最好使用tupledict创建变量,
    for i in range(10):
        for j in range(10):
            x[i, j] = model.addVar()
    # *********************************
    # *********************************
    # 添加约束
    #! 单个约束
    model.addConstr(x1+x2 == 1, name="c1")
    # 结合多维变量快速求和
    model.addConstr(x6.quicksum(1, "*", "*") == 1, name="c2")
    # 结合多维变量实现变量与权重的乘积求和:x3[0]*w[0] + x3[1]*w[1]
    model.addConstr(x3.quickprod({0: 1, 1: 2}, "*") >= 2)
    model.addConstr(model.sum([x4[i, j]
                    for i in range(2) for j in range(3)]) >= 0)
    #! 多约束
    model.addConstrs([x1+x2 == 1, x1+x2 >= 1, x1+x2 <= 1], name="c3")
    model.addConstrs(x3.quicksum(i, "*") <= 3 for i in range(2))
    #! 其他约束
    y = model.addVars(3, vtype=mipx.BINARY, name="y")
    model.addGenConstrAnd(y[0], [y[1], y[2]])  # and 运算 y[0] = min(y[1],y[2])
    model.addGenConstrOr(y[0], [y[1], y[2]])  # or 运算 y[0] = max(y[1],y[2])
    model.addGenConstrXOr(y[0], [y[1], y[2]])  # 异或运算
    #! 绝对值约束
    b = model.addVar(ub=10, name="b")
    model.addGenConstrAbs(b, x[3, 2]-x[2, 1]+x[1, 0], M=100)  # 不用目标函数辅助
    #! 乘积
    # b = x1*x2,其中x至少有一个0/1变量
    model.addConstrMultiply(b, (x1, x2))
    #! 条件约束
    z = model.addVar(vtype=mipx.BINARY, name="z")
    model.addGenConstrIndicator(z, False, y[0]+y[1], mipx.EQUAL, 1, M=100)
    # model.addIndicator(z, False, y[0]+y[1] == 1) # 待实现
    #! 约束满足的次数
    model.addConstrOr([x[0, 1]+x[1, 4] == 1, x[0, 3]-x[1, 2]
                       <= 1], ok_num=1, cmpType=mipx.EQUAL)
    #! 范围约束
    model.addRange(x[1, 2]+x[1, 3], 0, 10)

    #! 非重叠约束
    s1 = model.addVar(vtype=mipx.INTEGER, name="s1")
    e1 = model.addVar(vtype=mipx.INTEGER, name="e1")
    s2 = model.addVar(vtype=mipx.INTEGER, name="s2")
    e2 = model.addVar(vtype=mipx.INTEGER, name="e2")
    ivar = model.newIntervalVar(s1, 10, e1, name="interval")
    ivar2 = model.newIntervalVar(s2, 5, e2, name="interval2")
    model.addNoOverlap([ivar, ivar2], M=100)
    # *********************************
    # *********************************
    #! 其他辅助方法
    model.addKpi(x1+x2, "KPI1")
    model.addKpi(x1-x2, "KPI2")
    model.statistics()
    model.setNumThreads(2)  # 设置线程数
    model.setHint({x1: 1, x2: 0})  # 为变量设置初始值
    #! 设置优化目标
    #! 设置单目标
    # model.setObjective(x1+x2)
    #! 设置多目标 (与单目标二选一)
    model.setObjectiveN(x1+x2, index=0, weight=1, name="obj1")
    model.setObjectiveN(x1-x2, index=1, weight=1, name="obj2")
    model.setObjectiveN(x.quicksum(), index=2, weight=1, name="obj3")
    status = model.optimize(mipx.MINIMIZE)  # 优化 ,默认最小化
    # model.write("tst_model.mps")  # 支持后缀 .lp .mps
    if status == mipx.OptimizationStatus.OPTIMAL or status == mipx.OptimizationStatus.FEASIBLE:
        print("Optimal solution found.")
        print(f"耗时:{model.WallTime}ms")
        print(f"目标值:{model.ObjVal}")
        mipx.debugVar(y, no_zero=False)
        model.reportMultiObjValue()  # 输出多目标值
        # 输出kpi
        model.reportKpis()
        print("Kpi1:", model.kpiValueByName("KPI1"))  # 输出KPI1的值
        # 对应变量
        print("x:", x1.X)
        # 灵活的变量取值
        print("valueExpression:", model.valueExpression(x1+x[1, 2]))


def test_cpmodel():
    model = mipx.CpModel(name="test_cpmodel")
    #! 额外约束
    x = model.addVars(10, name='x')
    y = model.addVar(name='y')
    model.addGenConstrMin(y, [x[1], x[2], x[0]])  # y = min(x[1],x[2],x[0])
    model.addGenConstrMax(y, [x[3], x[4], x[5]])
    model.addGenConstrMultiplication(y, [x[4], x[6]])  # y = x[4]*x[6]
    index = model.addVar(ub=9, vtype=mipx.INTEGER, name='index_var')
    target = model.addVar(vtype=mipx.INTEGER, name='target_var')
    variables = model.addVars(
        10, lb=4, ub=7, vtype=mipx.INTEGER, name='variables')
    model.addElement(
        index=index, variables=variables.quickselect(), target=target)  # variables[index] = target
    m = model.addVar(lb=1, vtype=mipx.BINARY, name='m')
    n = model.addVar(vtype=mipx.BINARY, name='n')
    model.addImplication(m, n)  # 若m为真，则n一定为真。
    # ...
    model.addConstr(y >= 1)
    status = model.optimize()
    if status == mipx.OptimizationStatus.OPTIMAL or status == mipx.OptimizationStatus.FEASIBLE:
        print("Optimal solution found.")
        print(f"目标值:{model.ObjVal}")
        mipx.debugVar(x, no_zero=False)
        mipx.debugVar(y, no_zero=False)
        mipx.debugVar(index, no_zero=False)
        mipx.debugVar(target, no_zero=False)
        mipx.debugVar(m, no_zero=False)
        mipx.debugVar(n, no_zero=False)


if __name__ == '__main__':
    test_model()
    test_cpmodel()



# -*- coding: utf-8 -*-
# @Time    : 2023/4/28 23:54
# @Author  : luyi
from math import ceil, floor
from typing import Any

from .constants import Vtype
from .utilx import is_list_or_tuple
from .tupledict import tupledict

INFINITY_CP = 2**31 - 1
INFINITY = float("inf")


def get_update_lub(solver_id, is_cp, lb, ub: Any, vtype: Vtype):
    if is_cp and ub is None:
        ub = INFINITY_CP
    elif is_cp and ub:
        ...
    elif ub is None:
        ub = INFINITY
    if solver_id in ["SAT", "CBC"]:
        ub = INFINITY_CP if ub == INFINITY else ub
    if is_cp:
        if lb:
            lb = ceil(lb)
        if ub:
            ub = floor(ub)
    if vtype == Vtype.BINARY:
        if ub > 0 or ub is None:
            ub = 1
        if lb > 0:
            lb = 1
        if lb < 0 or lb is None:
            lb = 0
    ub = max(ub, 1)
    if lb > ub:
        raise Exception("0-1变量lb>=ub")
    return lb, ub


def abs_(*args):
    pass


def all_(*args):
    pass


def and_(*args):
    pass


def any_(*args):
    pass


def debugVar(vars, no_zero=True):
    """
    调试var的方法

    :param _type_ vars: 可以是变量的集合或者单个变量
    :param bool no_zero: 不输出<=0的值, defaults to True
    """
    is_list = False
    if type(vars) == type(tupledict()):
        vars = vars.values()
        is_list = True
    if is_list_or_tuple(vars):
        is_list = True
    if is_list:
        for var in vars:
            if no_zero:
                if var.X <= 0.01:
                    continue
            print(f"{var.VarName}:{var.X}")
    else:
        print(f"{vars.VarName}:{vars.X}")  # type:ignore



# -*- coding: utf-8 -*-
# @Time    : 2023/3/31 22:04
# @Author  : luyi
from typing import List, Optional, Union, Tuple, Dict

from ortools.linear_solver import pywraplp

try:
    from ortools.linear_solver.linear_solver_natural_api import OFFSET_KEY  # type: ignore
except Exception as e:
    from ortools.linear_solver.python.linear_solver_natural_api import OFFSET_KEY
from ortools.sat.python import cp_model
from .constants import Vtype, ObjType, OptimizationStatus, CmpType
from .interface_ import IModel, IVar, ITupledict
from .tupledict import tupledict
from .utilx import (
    is_list_or_tuple,
    get_combinations,
    name_str,
    is_generator,
    check_bool_var,
    is_bool_var,
)
from .variable import IntervalVar


INFINITY = float("inf")
INFINITY_CP = cp_model.INT32_MAX
Real = Union[float, int]


class LinearConstraint(pywraplp.LinearConstraint):
    """
    线性约束

    Args:
        pywraplp (_type_): 线性
    """

    def __init__(self, expr, lb, ub):
        self.__expr = expr
        self.__lb = lb
        self.__ub = ub
        super().__init__(expr, lb, ub)

    def lbAndUb(self) -> Tuple[float, float]:
        coeffs = self.__expr.GetCoeffs()
        constant = coeffs.pop(OFFSET_KEY, 0.0)
        lb = -INFINITY
        ub = INFINITY
        if self.__lb > -INFINITY:
            lb = self.__lb - constant
        if self.__ub < INFINITY:
            ub = self.__ub - constant
        return lb, ub

    def linear_expr(self) -> pywraplp.LinearExpr:
        return self.__expr


class CommonModelSolver(IModel):
    """
    模型
    """

    def __init__(self, solver_id="SCIP", name=""):
        super().__init__(solver_id)
        self.name = name
        self.__solver: pywraplp.Solver = pywraplp.Solver.CreateSolver(
            solver_id)
        self._flag_objective_n = False
        self._objective_n_list = []
        self._flag_objective = False
        self._num_inter_vars = 0
        self._num_binary_vars = 0
        self._num_coutinuous_vars = 0

    @property
    def Infinity(self) -> float:
        return INFINITY

    # @staticmethod
    def Sum(self, expr_array) -> float:
        result = pywraplp.SumArray(expr_array)
        return result  # type: ignore

    def setHint(self, start: Dict[IVar, Real]):
        _vars = []
        _values = []
        print(start)
        for var, value in start.items():
            _vars.append(var)
            _values.append(value)
        self.__solver.SetHint(_vars, _values)

    def setTimeLimit(self, time_limit_seconds):
        self.__solver.set_time_limit(time_limit_seconds * 1000)

    def wall_time(self) -> int:
        """
        求解所花费的时间 milliseconds
        :return:
        """
        return self.__solver.wall_time()

    def iterations(self) -> int:
        """迭代次数"""
        return self.__solver.iterations()

    def nodes(self) -> int:
        return self.__solver.nodes()

    def addVar(
        self,
        lb: Real = 0.0,
        ub: Real = INFINITY,
        vtype: Vtype = Vtype.CONTINUOUS,
        name: str = "",
    ) -> IVar:
        """
        Add a decision variable to a model.
        :param lb: Lower bound for new variable.
        :param ub: Upper bound for new variable.
        :param vtype: variable type for new variable(Vtype.CONTINUOUS, Vtype.BINARY, Vtype.INTEGER).
        :param name: Name for new variable.write
        :return: variable.
        """
        var: Optional[IVar] = None
        if vtype == Vtype.CONTINUOUS:
            self._num_coutinuous_vars += 1
            var = self.__solver.NumVar(lb, ub, name)
        elif vtype == Vtype.INTEGER:
            self._num_inter_vars += 1
            var = self.__solver.IntVar(lb, ub, name)
        elif vtype == Vtype.BINARY:
            self._num_binary_vars += 1
            var = self.__solver.IntVar(lb, ub, name)
        var.v_type = vtype  # type: ignore
        var._solver = self  # type: ignore
        return var  # type: ignore

    def addVars(
        self,
        *indices,
        lb: Real = 0.0,
        ub: Real = INFINITY,
        vtype: Vtype = Vtype.CONTINUOUS,
        name: str = ""
    ) -> ITupledict:
        li = []
        for ind in indices:
            if isinstance(ind, int):
                ind = [i for i in range(ind)]
            elif is_list_or_tuple(ind):
                pass
            elif isinstance(ind, str):
                ind = [ind]
            else:
                raise ValueError("error input")
            li.append(ind)
        all_keys_tuple = get_combinations(li)
        tu_dict = tupledict(
            [
                [key, self.addVar(lb, ub, vtype, name_str(name, key))]
                for key in all_keys_tuple
            ]
        )
        return tu_dict

    def getVars(self) -> List[IVar]:
        """
        Retrieve a list of all variables in the model.
        :return: All variables in the model.
        """
        return self.__solver.variables()

    def getVar(self, i) -> IVar:
        """
        获得第i个变量
        :param i:
        :return:
        """
        return self.__solver.variable(i)

    def addConstr(self, lin_expr, name="") -> LinearConstraint:  # type: ignore
        return self.__solver.Add(lin_expr, name)

    def addConstrs(self, lin_exprs, name=""):
        if not is_list_or_tuple(lin_exprs) and not is_generator(lin_exprs):
            raise RuntimeError("constraint conditions are not a set or list")
        for i, lin_expr in enumerate(lin_exprs):
            self.addConstr(lin_expr, name_str(name, i))

    def setObjective(self, expr):
        """
        Set the model objective equal to a linear expression
        :param expr:New objective expression
        :param obj_type:Optimization sense (Sense.MINIMIZE for minimization, Sense.MAXIMIZE for maximization)
        """
        self._flag_objective = True
        if self._flag_objective_n:
            raise RuntimeError(
                "setObjective and setObjectiveN can only be used for one of them"
            )
        # self.__solver.Minimize(expr)
        self._objective_n_list.append(expr)

    def _set_objective_sense(self, obj_type: ObjType):
        if obj_type == ObjType.MINIMIZE:
            self.__solver.Objective().SetMinimization()
        else:
            self.__solver.Objective().SetMaximization()

    def setObjectiveN(
        self, expr, index: int, priority: int = 0, weight: float = 1, name: str = ""
    ):
        """
        多目标优化，优化最小值
        :param name:
        :param expr: 表达式
        :param index: 目标函数对应的序号 (默认 0，1，2，…), 以 index=0 作为目标函数的值, 其余值需要另外设置参数
        :param priority:分层序列法多目标决策优先级(整数值), 值越大优先级越高,# 未实现。
        :param weight: 线性加权多目标决策权重(在优先级相同时发挥作用)
        """
        self._flag_objective_n = True
        if self._flag_objective:
            raise RuntimeError(
                "setObjective and setObjectiveN can only be used for one of them"
            )
        self._objective_n_list.append(expr*weight)
        # objective: pywraplp.Objective = self.__solver.Objective()
        # if isinstance(expr, (float, int)):
        #     objective.SetOffset(expr*weight)
        # else:
        #     coeffs = expr.GetCoeffs()
        #     objective.SetOffset(coeffs.pop(OFFSET_KEY, 0.0))
        #     for (
        #         v,
        #         c,
        #     ) in list(coeffs.items()):
        #         objective.SetCoefficient(v, float(c) * weight)

    def addGenConstrAnd(self, resvar, varList: List[IVar], name=""):
        """
        和 addGenConstrAnd(y, [x1,x2])
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        check_bool_var(resvar, varList)
        for var in varList:
            self.addConstr(resvar <= var)
        self.addConstr(resvar >= self.Sum(varList) - len(varList) + 1)

    def addGenConstrOr(self, resvar: IVar, varList: List[IVar], name=""):
        """
        或
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        check_bool_var(resvar, varList)
        for var in varList:
            self.addConstr(resvar >= var)
        self.addConstr(resvar <= self.Sum(varList))

    def addGenConstrXOr(self, resvar: IVar, varList: List[IVar], name=""):
        """
        异或
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        if len(varList) != 2:
            raise ValueError("length of vars must be 2")
        check_bool_var(resvar, varList)
        self.addConstr(resvar >= varList[0] - varList[1])
        self.addConstr(resvar >= varList[1] - varList[0])
        self.addConstr(resvar <= varList[0] + varList[1])
        self.addConstr(resvar <= 2 - varList[0] - varList[1])

    def addGenConstrPWL(self, var_y, var_x, x_range, y_value, name=""):
        """
        设置分段约束
        model.addGenConstrPWL(var, [1, 3, 5], [1, 2, 4])
        :param var_y: f(x)
        :param var_x:指定变量的目标函数是分段线性
        :param x:  定义分段线性变量的点的范围边界(非减序列)
        :param y:定义分段线性变量的范围所对应的目标函数的值
        :return:
        """
        raise NotImplementedError()

    def addGenConstrIndicator(
        self,
        binvar: IVar,
        binval: bool,
        lhs: IVar,
        sense: CmpType,
        rhs: float,
        M: float,
        name: str = "",
    ):
        """
        若 binvar 为binval ,则 lhs 与 rhs 之间有sense 的关系

        :param binvar: 0-1变量
        :param binval: bool 常量
        :param lhs:  左侧变量
        :param sense: 等号，大于等于，小于等于
        :param rhs: 右侧常量
        :param M: 大M
        :return:
        """
        if M is None:
            M = abs(rhs) + 100  # TODO:待优化
        if binval is True:
            z = 1 - binvar
        else:
            z = binvar
        if sense == CmpType.GREATER_EQUAL:
            self.addConstr(lhs + M * z >= rhs)
        elif sense == CmpType.EQUAL:
            self.addConstr(lhs + M * z >= rhs)
            self.addConstr(lhs - M * z <= rhs)
        else:
            self.addConstr(lhs - M * z <= rhs)

    def addIndicator(self, binvar: IVar, binval: bool, expr, name: str = ""):
        raise RuntimeError("not implemented")

    def addGenConstrAbs(self, resvar, var_abs: IVar, name="", M=None):  # type: ignore
        x = self.addVars(2, name="")
        _M = 10000000
        if M is not None:
            _M = M
        self.addConstr(var_abs == x[0] - x[1])
        self.addConstr(resvar == x[0] + x[1])
        y = self.addVar(vtype=Vtype.BINARY)
        # x[0] * x[1]==0 的约束
        self.addConstr(x[0] <= y * _M)
        self.addConstr(x[1] <= (1 - y) * _M)

    def addConstrMultiply(self, z: IVar, l: Tuple[IVar, IVar], name=""):
        """x * y = z"""
        super().addConstrMultiply(z, l)
        x = l[0]
        y = l[1]
        if not is_bool_var(x) and not is_bool_var(y):
            raise RuntimeError("At least one binary variable is required.")
        if is_bool_var(y):
            x, y = y, x
        M = y.Ub
        if self.solver_id in ["SAT", "CBC"]:
            M = INFINITY_CP if M == INFINITY else M
        self.addConstr(z <= y)
        self.addConstr(z <= x * M)
        self.addConstr(z >= y + (x - 1) * M)

    def addRange(
        self, expr, min_value: Union[float, int], max_value: Union[float, int], name=""
    ):
        """
        添加范围约束
        :param
        expr:
        :param
        min_value:
        :param
        max_value:
        :param
        name:
        :return:
        """
        if isinstance(min_value, (int, float)) and isinstance(max_value, (int, float)):
            if min_value > max_value:
                raise ValueError("min_value is bigger than max_value")
        self.addConstr(expr >= min_value)
        self.addConstr(expr <= max_value)

    def addConstrOr(
        self,
        constrs: List[pywraplp.LinearConstraint],
        ok_num: int = 1,
        cmpType: CmpType = CmpType.EQUAL,
        name: str = "",
        M=None,
    ):
        """
            约束的满足情况

        :param constr: 所有的约束
        :param ok_num: 需要满足的个数，具体则根据cmpType
        :param cmpType: CmpType.LESS_EQUAL CmpType.EQUAL,CmpType.GREATER_EQUAL
        :param name: 名称
        :return:
        """
        constr_num = len(constrs)
        x = []
        for i in range(constr_num):
            x.append(self.addVar(vtype=Vtype.BINARY))
            tempConstr = _change_to_linear_constraint(constrs[i])
            lb, ub = tempConstr.lbAndUb()
            expr = tempConstr.linear_expr()
            tempM = 0
            if M is not None:
                tempM = M
            if lb > -INFINITY:  # 若大于
                if M is None:
                    tempM = (abs(lb) + 1) * 10
                self.addConstr(expr + tempM * (1 - x[i]) >= lb)
            if ub < INFINITY:
                if M is None:
                    tempM = (abs(ub) + 1) * 10
                self.addConstr(expr - tempM * (1 - x[i]) <= ub)
        if cmpType == CmpType.EQUAL:
            self.addConstr(self.Sum(x) == ok_num)
        elif cmpType == CmpType.GREATER_EQUAL:
            self.addConstr(self.Sum(x) >= ok_num)
        elif cmpType == CmpType.LESS_EQUAL:
            self.addConstr(self.Sum(x) <= ok_num)  # type: ignore
        else:
            raise Exception("error value of cmpType")

    def numVars(self, vtype: Optional[Vtype] = None) -> int:
        """
        当前的变量的个数
        :return:
        """
        if vtype is Vtype.CONTINUOUS:
            return self._num_coutinuous_vars
        elif vtype is Vtype.INTEGER:
            return self._num_inter_vars
        elif vtype is Vtype.BINARY:
            return self._num_binary_vars
        else:
            return self.__solver.NumVariables()

    def numConstraints(self) -> int:
        """
        当前的约束的个数
        :return:
        """
        return self.__solver.NumConstraints()

    def write(self, filename: str, obfuscated=False):
        """
        写入到文件
        :param filename:文件名，支持后缀 .lp .mps .proto
        :param obfuscated:
        :return:
        """
        filename = filename.lower()
        content = ""
        if filename.endswith(".lp"):
            content = self.__solver.ExportModelAsLpFormat(False)
        elif filename.endswith(".mps"):
            content = self.__solver.ExportModelAsMpsFormat(True, obfuscated)
        elif filename.endswith(".proto"):
            raise TypeError(".proto 导出异常，待修复")

        with open(filename, "w", encoding="utf-8") as f:
            f.write(content)

    def read(self, path: str):
        raise RuntimeError("not implemented")

    def getVarByName(self, name: str) -> IVar:
        return self.__solver.LookupVariable(name)

    @property
    def ObjVal(self):
        return self.__solver.Objective().Value()

    def _set_params(self):
        if self.Params.TimeLimit:
            self.setTimeLimit(self.Params.TimeLimit)
            # 是否允许输出运算信息，包括gap等
        if self.Params.EnableOutput:
            self.__solver.EnableOutput()

    def optimize(self, obj_type: ObjType = ObjType.MINIMIZE) -> OptimizationStatus:
        self._set_params()
        # GAP 设置。
        params = pywraplp.MPSolverParameters()
        # 配置参数
        if self.Params.MIPGap:
            params.SetDoubleParam(params.RELATIVE_MIP_GAP, self.Params.MIPGap)
        if self.Params.Precision:
            params.SetDoubleParam(params.PRIMAL_TOLERANCE,
                                  self.Params.Precision)
        if obj_type == ObjType.MINIMIZE:
            self.__solver.Minimize(self.Sum(self._objective_n_list))
        else:
            self.__solver.Maximize(self.Sum(self._objective_n_list))
        status = self.__solver.Solve(params)

        if status == pywraplp.Solver.OPTIMAL:
            result = OptimizationStatus.OPTIMAL
        elif status == pywraplp.Solver.INFEASIBLE:
            result = OptimizationStatus.INFEASIBLE
        elif status == pywraplp.Solver.UNBOUNDED:
            result = OptimizationStatus.UNBOUNDED
        elif status == pywraplp.Solver.FEASIBLE:
            result = OptimizationStatus.FEASIBLE
        elif status == pywraplp.Solver.NOT_SOLVED:
            result = OptimizationStatus.NO_SOLUTION_FOUND
        else:
            result = OptimizationStatus.ERROR
        return result

    def clear(self):
        self.__solver.Clear()

    def close(self):
        self.clear()
        self.__solver = None  # type: ignore

    def valueExpression(self, expression):
        return expression.solution_value()
        # value = 0
        # coeffs = expression.GetCoeffs()
        # for key, coeff in coeffs.items():
        #     if key is OFFSET_KEY:
        #         value += coeffs.get(key)
        #     else:
        #         var: IVar = key
        #         x = var.X
        #         value += coeff * x
        # return value

    def newIntervalVar(self, start, size, end, name="") -> IntervalVar:  # type: ignore
        self.addConstr(start + size == end, name=name)
        return IntervalVar(start, size, end)

    def addNoOverlap(self, interval_vars: List[IntervalVar], M: int):
        m = INFINITY_CP if M is None else M
        for i, var_i in enumerate(interval_vars):
            for j, var_j in enumerate(interval_vars):
                if i == j:
                    continue
                t = self.addVar(vtype=Vtype.BINARY)
                self.addConstr(var_i.start >= var_j.end - (1 - t) * m)
                self.addConstr(var_j.start >= var_i.end - t * m)

    def setNumThreads(self, num_theads: int):
        self.__solver.SetNumThreads(num_theads)


def _change_to_linear_constraint(
    constraint: pywraplp.LinearConstraint,
) -> LinearConstraint:
    expr = constraint._LinearConstraint__expr  # type: ignore
    ub = constraint._LinearConstraint__ub  # type: ignore
    lb = constraint._LinearConstraint__lb  # type: ignore
    return LinearConstraint(expr, lb, ub)



# -*- coding: utf-8 -*-
# @Time    : 2023/4/28 15:33
# @Author  : luyi
# 定义接口。统一类型。
from abc import ABCMeta, abstractmethod
from collections import UserDict, UserList
import sys
from typing import Any, List, Literal, Optional, Union, Dict, Generator, Tuple

from .constants import Vtype, ObjType, CmpType, OptimizationStatus, Params

INFINITY = float("inf")

Real = Union[float, int]


class IVar:
    def __init__(self):
        # 变量类型
        self.v_type = None
        self._solver = None

    @property
    def X(self) -> Real: ...

    @property
    def VarIndex(self) -> Real: ...

    @property
    def VarName(self) -> str: ...

    @property
    def Lb(self) -> Real: ...

    @property
    def Ub(self) -> Real: ...

    def setValue(self, value: Real): ...

    def setUb(self, ub: Real): ...

    def setLb(self, lb: Real): ...

    def Not(self) -> "IVar":
        """
        返回一个相反的变量，仅用于0,1变量。
        :return:
        """
        ...

    def setBounds(self, lb, ub): ...

    def __add__(self, expr) -> Any: ...

    def __radd__(self, cst) -> Any: ...

    def __sub__(self, expr) -> Any: ...

    def __rsub__(self, cst) -> Any: ...

    def __mul__(self, other) -> Any: ...

    def __rmul__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __le__(self, other) -> Any: ...


class ITupledict(UserDict):
    """
    Custom Gurobi class: tupledict is a subclass of dict where
      the keys are a tuplelist.
    """

    @abstractmethod
    def prod(self, coeff: Dict, *pattern): ...

    @abstractmethod
    def select(self, *pattern) -> List: ...

    @abstractmethod
    def sum(self, *pattern) -> IVar: ...

    @abstractmethod
    def quicksum(self, *pattern) -> IVar: ...

    @abstractmethod
    def quickselect(self, *pattern) -> List: ...

    @abstractmethod
    def quickprod(self, coeff: Dict, *pattern) -> IVar: ...


class ITuplelist(UserList):
    @abstractmethod
    def select(self, *args) -> List: ...


class IModel(metaclass=ABCMeta):
    def __init__(
        self,
        solver_id: Literal[
            "SCIP",
            "CBC",
            "CP",
            "SAT",
            "CLP",
            "CPLEX",
            "CPLEX_CP",
            "GLOP",
            "BOP",
            "GLPK",
        ],
        name="",
    ):
        self.Params: Params = Params()
        self.solver_id = solver_id

    @property
    @abstractmethod
    def Infinity(self) -> Real: ...

    @abstractmethod
    def Sum(self, expr_array) -> Real: ...

    @abstractmethod
    def setHint(self, start: Dict[IVar, Real]): ...

    @abstractmethod
    def setTimeLimit(self, time_limit_seconds: int):
        """
        设置程序最大运行时间
        :param time_limit_seconds: 秒
        """
        ...

    @abstractmethod
    def wall_time(self) -> Real:
        """
        求解所花费的时间
        :return: 求解所花费的时间，单位毫秒
        """
        ...

    @abstractmethod
    def iterations(self) -> Real:
        """

        :return: 算法迭代的次数
        """
        ...

    @abstractmethod
    def nodes(self) -> Real:
        """

        :return: 节点数
        """
        ...

    @abstractmethod
    def addVar(
        self,
        lb: float = 0,
        ub: float = INFINITY,
        vtype: Vtype = Vtype.CONTINUOUS,
        name: str = "",
    ) -> IVar:
        """
        创建变量
        :param lb: 变量下界
        :param ub: 变量上界
        :param vtype: 变量类型： Vtype.CONTINUOUS（连续）,Vtype.BINARY(0-1变量), Vtype.BINARY（整数变量）
        :param name:变量名
        :return:变量实体
        """
        ...

    @abstractmethod
    def addVars(
        self,
        *indices,
        lb: float = 0,
        ub: float = INFINITY,
        vtype: Vtype = Vtype.CONTINUOUS,
        name: str = "",
    ) -> ITupledict:
        """
        创建多维变量
        :param indices:多维的参数，如addVars(1,2),addVars(mList,nList),addVars([1,2,3],[3,4,5])等。
        :param lb:变量下界
        :param ub:变量上界
        :param vtype:变量类型： Vtype.CONTINUOUS（连续）,Vtype.BINARY(0-1变量), Vtype.BINARY（整数变量）
        :param name:变量名
        :return:tupledict类型
        """
        ...

    @abstractmethod
    def getVars(self) -> List[IVar]:
        """
        获取所有的变量对象
        :return:
        """
        ...

    @abstractmethod
    def getVar(self, i) -> IVar:
        """
        获取第i个变量
        :param i:
        :return:
        """
        ...

    @abstractmethod
    def addConstr(self, lin_expr, name=""):
        """
        向模型添加约束条件，
        :param lin_expr:线性约束表达式
        :param name: 约束名称
        :return:
        """
        ...

    @abstractmethod
    def addConstrs(self, lin_exprs: Union[List, Generator, Tuple], name=""):
        """
        向模型添加多个约束条件，
        :param lin_exprs: 线性约束表达式集合 可以为列表或者元组。
        :param name:名称
        :return:
        """
        ...

    @abstractmethod
    def setObjective(self, expr):
        """
        设置模型的单一目标
        :param expr: 目标表达式
        优化方向。ObjType.MINIMIZE（最小值），ObjType.MAXIMIZE(最大值)
        :return:
        """
        ...

    @abstractmethod
    def setObjectiveN(
        self, expr, index: int, priority: int = 0, weight: float = 1, name: str = ""
    ):
        """
        多目标优化，优化最小值
        :param expr: 表达式
        :param index: 目标函数对应的序号 (默认 0，1，2，…), 以 index=0 作为目标函数的值, 其余值需要另外设置参数
        :param priority: 分层序列法多目标决策优先级(整数值), 值越大优先级越高【未实现】
        :param weight: 线性加权多目标决策权重(在优先级相同时发挥作用)
        :param name: 名称
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrAnd(self, resvar, varList: List[IVar], name=""):
        """
        and 运算。
        addGenConstrAnd(y, [x1, x2]) 表示y = max(x1,x2)。 所有变量均为0-1变量
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrOr(self, resvar: IVar, varList: List[IVar], name=""):
        """
        或运算
        addGenConstrOr(y, [x1, x2]) 表示y = min(x1,x2)。 所有变量均为0-1变量
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrXOr(self, resvar: IVar, varList: List[IVar], name=""):
        """
        异或运算
        addGenConstrXOr(y, [x1, x2])。 所有变量均为0-1变量
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrPWL(
        self,
        var_y: IVar,
        var_x: IVar,
        x_range: List[float],
        y_value: List[float],
        name="",
    ):
        """
        设置分段约束
        model.addGenConstrPWL(var, [1, 3, 5], [1, 2, 4])
        :param var_y: f(x)
        :param var_x:指定变量的目标函数是分段线性
        :param x:  定义分段线性目标函数的点的横坐标值(非减序列)
        :param y:定义分段线性目标函数的点的纵坐标值
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrIndicator(
        self,
        binvar: IVar,
        binval: bool,
        lhs: IVar,
        sense: CmpType,
        rhs: float,
        M: float,
        name: str = "",
    ):
        """
        若 binvar 为binval ,则 lhs 与 rhs 之间有sense 的关系
        若M不指定,则程序会给与默认。但仍推荐给出M。程序自动给出的可能会存在问题。
        :param binvar: 0-1变量
        :param binval: bool 常量
        :param lhs:  左侧变量
        :param sense: 等号，大于等于，小于等于
        :param rhs: 右侧常量
        :param M: 大M
        :return:
        """
        ...

    @abstractmethod
    def addIndicator(self, binvar: IVar, binval: bool, expr, name: str = ""):
        """
        若 binvar 为binval ,则 lhs 与 rhs 之间有sense 的关系
        若M不指定，则程序会给与默认。但仍推荐给出M。程序自动给出的可能会存在问题。
        :param binvar: 0-1变量
        :param binval: bool 常量
        :param lhs:  左侧变量
        :param sense: 等号，大于等于，小于等于
        :param rhs: 右侧常量
        :param M: 大M
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrAbs(self, resvar: any, var_abs: IVar, M, name=""):  # type: ignore
        """
        绝对值 resvar = |var_abs|
        :param resvar:
        :param var_abs:
        :param name:
        :return:
        """
        ...

    @abstractmethod
    def addConstrMultiply(self, z: IVar, l: Tuple[IVar, IVar], name=""):
        """
        满足 z = x * y
        其中 x 为0,1变量
        :param l:
        :param z: 变量
        :param name:
        :return:
        """
        if len(list(l)) != 2:
            raise RuntimeError("Only need two variables")

    @abstractmethod
    def addRange(self, expr, min_value: Real, max_value: Real, name=""):
        """
        添加范围约束
        :param expr: 表达式
        :param min_value: 最小值
        :param max_value: 最大值
        :param name:名称
        :return:
        """
        ...

    @abstractmethod
    def addConstrOr(
        self,
        constrs: List,
        ok_num: int = 1,
        cmpType: CmpType = CmpType.EQUAL,
        name: str = "",
        M=None,
    ):
        """
        约束的满足情况,满足的次数
        :param constr: 所有的约束
        :param ok_num:  需要满足的个数，具体则根据cmpType
        :param cmpType: CmpType.LESS_EQUAL CmpType.EQUAL,CmpType.GREATER_EQUAL
        :param name: 名称
        :param M: M值，推荐指定M值。
        :return:
        """
        ...

    @abstractmethod
    def numVars(self, vtype: Optional[Vtype] = None) -> int:
        """
        变量个数
        :return:
        """
        ...

    @abstractmethod
    def numConstraints(self) -> int:
        """
        约束个数
        :return:
        """
        ...

    @abstractmethod
    def write(self, filename: str, obfuscated=False):
        """
        写入到文件
        :param filename:文件名，支持后缀 .lp .mps .proto(目前有问题)
        :param obfuscated: 是否混淆，默认不混淆
        :return:
        """
        ...

    @abstractmethod
    def read(self, path: str):
        """读取文件 lp,mps 文件。

        Args:
            path (str): _description_
        """
        ...

    @property
    @abstractmethod
    def ObjVal(self) -> Real:
        """目标值"""
        ...

    @abstractmethod
    def optimize(self, obj_type: ObjType = ObjType.MINIMIZE) -> OptimizationStatus:
        """
        优化目标
        :param time_limit_milliseconds:最大运行时长
        :param obj_type:优化目标。ObjType.MINIMIZE（最小值），ObjType.MAXIMIZE(最大值)
        :param enable_output: 是否显示gap日志。
        :return:
        """
        ...

    @abstractmethod
    def clear(self):
        """清空所有的目标和约束"""
        ...

    @abstractmethod
    def close(self):
        """关闭"""
        ...

    @abstractmethod
    def valueExpression(self, expression):
        """
        计算表达式的值。
        :param expression:
        :return:
        """
        ...

    @abstractmethod
    def newIntervalVar(self, start, size, end, name=""):
        """
        创建变量： start+size=end

        Args:
            start (_type_): 开始
            size (_type_): 大小
            end (_type_): 结束
            name (str, optional): 名称. Defaults to "".
        """
        ...

    @abstractmethod
    def addNoOverlap(self, interval_vars: List, M):
        """
        互相之间不重复

        Args:
            interval_vars (List): 间隔变量
        """
        pass

    @abstractmethod
    def setNumThreads(self, num_theads: int):
        """
        设置线程的个数

        Args:
            num_theads (int): 线程个数
        """
        ...


class ICpModel(metaclass=ABCMeta):
    def __init__(self, solver_id: Literal["CPLEX_CP", "CP"], name=""):
        self.Params: Params = Params()
        self.solver_id = solver_id

    @property
    @abstractmethod
    def Infinity(self) -> Real: ...

    @staticmethod
    @abstractmethod
    def Sum(expr_array) -> Any: ...

    @abstractmethod
    def setHint(self, start: Dict[IVar, Real]): ...

    @abstractmethod
    def setTimeLimit(self, time_limit_seconds: int):
        """
        设置程序最大运行时间
        :param time_limit_seconds: 秒
        """
        ...

    @abstractmethod
    def wall_time(self) -> Real:
        """
        求解所花费的时间
        :return: 求解所花费的时间，单位毫秒
        """
        ...

    @abstractmethod
    def iterations(self) -> Real:
        """

        :return: 算法迭代的次数
        """
        ...

    @abstractmethod
    def nodes(self) -> Real:
        """

        :return: 节点数
        """
        ...

    @abstractmethod
    def addVar(
        self,
        lb: int = 0,
        ub: int = sys.maxsize,
        vtype: Literal[Vtype.BINARY, Vtype.INTEGER] = Vtype.INTEGER,
        name: str = "",
    ) -> IVar:
        """
        创建变量
        :param lb: 变量下界
        :param ub: 变量上界
        :param vtype: 变量类型： Vtype.CONTINUOUS（连续）,Vtype.BINARY(0-1变量), Vtype.BINARY（整数变量）
        :param name:变量名
        :return:变量实体
        """
        ...

    @abstractmethod
    def addVars(
        self,
        *indices,
        lb: int = 0,
        ub: int = sys.maxsize,
        vtype: Literal[Vtype.BINARY, Vtype.INTEGER] = Vtype.INTEGER,
        name: str = "",
    ) -> ITupledict:
        """
        创建多维变量
        :param indices:多维的参数，如addVars(1,2),addVars(mList,nList),addVars([1,2,3],[3,4,5])等。
        :param lb:变量下界
        :param ub:变量上界
        :param vtype:变量类型： Vtype.CONTINUOUS（连续）,Vtype.BINARY(0-1变量), Vtype.BINARY（整数变量）
        :param name:变量名
        :return:tupledict类型
        """
        ...

    @abstractmethod
    def getVars(self) -> List[IVar]:
        """
        获取所有的变量对象
        :return:
        """
        ...

    @abstractmethod
    def getVar(self, i) -> IVar:
        """
        获取第i个变量
        :param i:
        :return:
        """
        ...

    @abstractmethod
    def addConstr(self, lin_expr, name=""):
        """
        向模型添加约束条件，
        :param lin_expr:线性约束表达式
        :param name: 约束名称
        :return:
        """
        ...

    @abstractmethod
    def addConstrs(self, lin_exprs: Union[List, Generator, Tuple], name=""):
        """
        向模型添加多个约束条件，
        :param lin_exprs: 线性约束表达式集合 可以为列表或者元组。
        :param name:名称
        :return:
        """
        ...

    @abstractmethod
    def setObjective(self, expr):
        """
        设置模型的单一目标
        :param expr: 目标表达式
        优化方向。ObjType.MINIMIZE（最小值），ObjType.MAXIMIZE(最大值)
        :return:
        """
        ...

    @abstractmethod
    def setObjectiveN(
        self, expr, index: int, priority: int = 0, weight: float = 1, name: str = ""
    ):
        """
        多目标优化，优化最小值
        :param expr: 表达式
        :param index: 目标函数对应的序号 (默认 0，1，2，…), 以 index=0 作为目标函数的值, 其余值需要另外设置参数
        :param priority: 分层序列法多目标决策优先级(整数值), 值越大优先级越高【未实现】
        :param weight: 线性加权多目标决策权重(在优先级相同时发挥作用)
        :param name: 名称
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrAnd(self, resvar, varList: List[IVar], name=""):
        """
        and 运算。
        addGenConstrAnd(y, [x1, x2]) 表示y = max(x1,x2)。 所有变量均为0-1变量
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrOr(self, resvar: IVar, varList: List[IVar], name=""):
        """
        或运算
        addGenConstrOr(y, [x1, x2]) 表示y = min(x1,x2)。 所有变量均为0-1变量
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrXOr(self, resvar: IVar, varList: List[IVar], name=""):
        """
        异或运算
        addGenConstrXOr(y, [x1, x2])。 所有变量均为0-1变量
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrPWL(
        self,
        var_y: IVar,
        var_x: IVar,
        x_range: List[int],
        y_value: List[int],
        name="",
    ):
        """
        设置分段约束
        model.addGenConstrPWL(var, [1, 3, 5], [1, 2, 4])
        :param var_y: f(x)
        :param var_x:指定变量的目标函数是分段线性
        :param x:  定义分段线性目标函数的点的横坐标值(非减序列)
        :param y:定义分段线性目标函数的点的纵坐标值
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrIndicator(
        self,
        binvar: IVar,
        binval: bool,
        lhs: IVar,
        sense: CmpType,
        rhs: int,
        M: int,
        name: str = "",
    ):
        """
        若 binvar 为binval ,则 lhs 与 rhs 之间有sense 的关系
        若M不指定,则程序会给与默认。但仍推荐给出M。程序自动给出的可能会存在问题。
        :param binvar: 0-1变量
        :param binval: bool 常量
        :param lhs:  左侧变量
        :param sense: 等号，大于等于，小于等于
        :param rhs: 右侧常量
        :param M: 大M
        :return:
        """
        ...

    @abstractmethod
    def addIndicator(self, binvar: IVar, binval: bool, expr, name: str = ""):
        """
        若 binvar 为binval ,则 lhs 与 rhs 之间有sense 的关系
        若M不指定，则程序会给与默认。但仍推荐给出M。程序自动给出的可能会存在问题。
        :param binvar: 0-1变量
        :param binval: bool 常量
        :param lhs:  左侧变量
        :param sense: 等号，大于等于，小于等于
        :param rhs: 右侧常量
        :param M: 大M
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrAbs(self, resvar: any, var_abs: IVar):  # type: ignore
        """
        绝对值 resvar = |var_abs|
        :param resvar:
        :param var_abs:
        :param name:
        :return:
        """
        ...

    @abstractmethod
    def addConstrMultiply(self, z: IVar, l: Tuple[IVar, IVar], name=""):
        """
        满足 z = x * y
        其中 x 为0,1变量
        :param l:
        :param z: 变量
        :param name:
        :return:
        """
        if len(list(l)) != 2:
            raise RuntimeError("Only need two variables")

    @abstractmethod
    def addRange(self, expr, min_value: Real, max_value: Real, name=""):
        """
        添加范围约束
        :param expr: 表达式
        :param min_value: 最小值
        :param max_value: 最大值
        :param name:名称
        :return:
        """
        ...

    @abstractmethod
    def addConstrOr(
        self,
        constrs: List,
        ok_num: int = 1,
        cmpType: CmpType = CmpType.EQUAL,
        name: str = "",
        M=None,
    ):
        """
        约束的满足情况,满足的次数
        :param constr: 所有的约束
        :param ok_num:  需要满足的个数，具体则根据cmpType
        :param cmpType: CmpType.LESS_EQUAL CmpType.EQUAL,CmpType.GREATER_EQUAL
        :param name: 名称
        :param M: M值，推荐指定M值。
        :return:
        """
        ...

    @abstractmethod
    def numVars(self, vtype: Optional[Vtype] = None) -> int:
        """
        变量个数
        :return:
        """
        ...

    @abstractmethod
    def numConstraints(self) -> int:
        """
        约束个数
        :return:
        """
        ...

    @abstractmethod
    def write(self, filename: str, obfuscated=False):
        """
        写入到文件
        :param filename:文件名，支持后缀 .lp .mps .proto(目前有问题)
        :param obfuscated: 是否混淆，默认不混淆
        :return:
        """
        ...

    @abstractmethod
    def read(self, path: str) -> "IModel":
        """读取文件 lp,mps 文件。

        Args:
            path (str): _description_
        """
        ...

    @property
    @abstractmethod
    def ObjVal(self) -> Real:
        """目标值"""
        ...

    @abstractmethod
    def optimize(self, obj_type: ObjType = ObjType.MINIMIZE) -> OptimizationStatus:
        """
        优化目标
        :param time_limit_milliseconds:最大运行时长
        :param obj_type:优化目标。ObjType.MINIMIZE（最小值），ObjType.MAXIMIZE(最大值)
        :param enable_output: 是否显示gap日志。
        :return:
        """
        ...

    @abstractmethod
    def clear(self):
        """清空所有的目标和约束"""
        ...

    @abstractmethod
    def close(self):
        """关闭"""
        ...

    @abstractmethod
    def valueExpression(self, expression) -> int:
        """
        计算表达式的值。
        :param expression:
        :return:
        """
        ...

    @abstractmethod
    def newIntervalVar(self, start, size, end, name="") -> Any:
        """
        创建变量： start+size=end

        Args:
            start (_type_): 开始
            size (_type_): 大小
            end (_type_): 结束
            name (str, optional): 名称. Defaults to "".
        """
        ...

    @abstractmethod
    def addNoOverlap(self, interval_vars: List, M):
        """
        互相之间不重复

        Args:
            interval_vars (List): 间隔变量
        """
        pass

    @abstractmethod
    def setNumThreads(self, num_theads: int):
        """
        设置线程的个数

        Args:
            num_theads (int): 线程个数
        """
        ...

    @abstractmethod
    def addElement(self, index, variables, target): ...

    @abstractmethod
    def addCircuit(self, arcs): ...

    @abstractmethod
    def addAllowedAssignments(self, variables, tuples_list): ...

    @abstractmethod
    def addForbiddenAssignments(self, variables, tuples_list): ...

    @abstractmethod
    def addInverse(
        self,
        variables,
        inverse_variables,
    ): ...

    @abstractmethod
    def addMapDomain(self, var, bool_var_array, offset):
        """Adds `var == i + offset <=> bool_var_array[i] == true for all i`."""
        ...

    @abstractmethod
    def addImplication(self, a, b): ...

    @abstractmethod
    def addBoolTrueOr(self, literals): ...

    @abstractmethod
    def addBoolTrueAnd(self, literals): ...

    @abstractmethod
    def addAtLeastOneIsTrue(self, literals): ...

    @abstractmethod
    def addAtMostOneIsTrue(self, literals): ...

    @abstractmethod
    def addExactlyNumIsTrue(self, literals, num: int = 1): ...

    @abstractmethod
    def addGenConstrMax(self, target: IVar, varList: List[IVar]):
        """
        最大值约束
        """
        ...

    @abstractmethod
    def addGenConstrMin(self, target: IVar, varList: List[IVar]):
        """
        最小值约束
        """
        ...

    @abstractmethod
    def addGenConstrDivision(self, target, num, denom):
        """Adds `target == num // denom` (integer division rounded towards 0)."""
        ...

    @abstractmethod
    def addGenConstrModulo(self, target, num, denom):
        """Adds `target = expr % mod`."""
        ...

    @abstractmethod
    def addGenConstrMultiplication(
        self,
        target,
        *expressions,
    ): ...


def debugVar(vars, no_zero=True):
    """
    调试var的方法

    :param _type_ vars: 可以是变量的集合或者单个变量
    :param bool no_zero: 不输出<=0的值, defaults to True
    """
    ...



# -*- coding: utf-8 -*-
# @Time    : 2023/4/1 10:41
# @Author  : luyi
from collections import UserList
from typing import List, Iterable
from warnings import warn

from .interface_ import ITuplelist
from .utilx import get_length, is_list_or_tuple


class tuplelist(ITuplelist):
    """
    Custom  class: tuplelist is a subclass of list that is
      designed to work with lists of tuples.  Using the select()
      method, this class allows you to efficiently select sub-lists of
      tuples by matching specific values in specific fields of the
      member tuples.

      For example:
        > l = tuplelist([(1, 2), (1, 3), (2, 3), (2, 4)])
        > print(l.select('*', '*'))
        [(1, 2), (1, 3), (2, 3), (2, 4)]
        > print(l.select('*', 3))
        [(1, 3), (2, 3)]
        > print(l.select(1, '*'))
        [(1, 2), (1, 3)]

      A tuplelist is designed to store tuples containing scalar values (int,
      float, string, ...). It may produce unpredictable results with other
      Python objects, such as tuples of tuples. Thus, you can store
      (1, 2.0, 'abc') in a tuplelist, but you shouldn't store ((1, 2.0), 'abc').
    """

    def append(self, item):
        # 检查是否重复。若重复则不添加。
        if isinstance(item, UserList) or type(item) == type([]):  # 由于特殊性，不进行过多的扩展。
            raise ValueError("append传入参数不能为集合")
        if item in self.data:
            return
        # 检查维度
        if self.dim is None:
            self.dim = get_length(item)
        else:
            if self.dim != get_length(item):
                raise ValueError("传入的数据维度不一致。")
        super().append(item)

    def extend(self, other: Iterable[any]):  # type: ignore
        # 检查是否重复。
        if isinstance(other, UserList):
            datas = other.data
        else:
            datas = other
        for item in datas:
            self.append(item)

    def __init__(self, seq=None):
        seq = self._duplicate(seq)
        if len(seq) == 0:
            self.dim = None
            super().__init__([])
        else:
            self.dim = get_length(seq[0])
            super().__init__(seq)

    def _get_dim(self):
        if self.dim is None:
            if len(self) == 0:
                self.dim = 0
                return self.dim
            self.dim = get_length(self[0])
        return self.dim

    def quickselect(self, *pattern) -> List:
        dim = len(pattern)
        if dim == 0:
            return list(self)  # type: ignore
        if len(self) == 0:
            return []
        li = []
        if self._get_dim() == 1:
            for key in self:
                p = pattern[0]
                if p == "*" or p == key:
                    li.append(key)
            return li
        for key in self:
            ok = True
            # 判定key是否可以迭代
            for i, k in enumerate(key):
                p = pattern[i]
                if p == "*":
                    continue
                if p != k:
                    ok = False
                    break
            if ok:
                li.append(key)
        return li

    def select(self, *args) -> List:
        warn(f"call to deprecated function,use quickselect function")
        values = [*args]
        if len(self.data) == 0:
            return []
        if len(values) == 0:  # 全选
            return self.data
        if self.dim is None:
            self.dim = len(values)
        if self.dim != len(values):
            raise ValueError("传入的数据个数异常")
        _dim_ok_index = []
        _list_num_range = [i for i in range(len(self.data))]
        # 为了减少其他方法如pop,remove 等操作。将计算的操作放入该位置。
        # 会存在重复计算的问题。
        _keys = []
        for i in range(self.dim):
            temp = []
            for tu in self.data:
                _dim = get_length(tu)
                # 检查传入的参数维度是否一致
                if _dim != self.dim:
                    raise ValueError("传入的数据维度不一致。")
                if is_list_or_tuple(tu):
                    temp.append(tu[i])
                else:
                    temp.append(tu)
            _keys.append(temp)

        for i, key in enumerate(values):
            if key == "*":
                _dim_ok_index.append(_list_num_range)
                continue
            if isinstance(key, (float, int, str)):
                key = [key]
            _dim_ok_index.append(
                [j for j, x in enumerate(_keys[i]) if x in key])
        flag = []
        for i, item in enumerate(_dim_ok_index):
            if i == 0:
                flag = item
                continue
            flag = list(set(flag).intersection(set(item)))
        return [key for i, key in enumerate(self.data) if i in flag]

    @staticmethod
    def _duplicate(initList):
        data = []
        if initList is not None:
            if type(initList) == type(data):
                data[:] = initList
            elif isinstance(initList, UserList):
                data[:] = initList.data[:]
            else:
                data = list(initList)
        l = [key for key in data]
        dim = None  # 检查传入的数据是否都是同一纬度的。
        for i, key in enumerate(l):
            if isinstance(key, list):
                t_key = tuple(key)
                l[i] = t_key
            _dim = get_length(key)
            if dim is None:
                dim = _dim
            else:
                if dim != _dim:
                    raise ValueError("传入的数据维度不一致。")
        return list(set(l))


if __name__ == '__main__':
    a = tuplelist([(1, 2), (1, 2), [2, 3]])
    print(a.quickselect())
    print({1, 1})




import mipx


if __name__ == "__main__":
    model = mipx.Model("CPLEX")
    x = model.addVar(lb=1, ub=10, name='x')
    y = model.addVar(lb=1, ub=10, name='y')
    # model.setObjective(10)
    # model.setObjectiveN(10*x, index=0, weight=10)
    # model.setObjectiveN(10*x, index=1, weight=10)
    # model.setObjectiveN(10*y, index=3, weight=1)
    # model.setObjectiveN(x, index=2, weight=1)
    # model.setObjectiveN(100, index=4, weight=1)
    model.setObjectiveN(100, index=4, weight=1)
    model.setObjective(10*x + 10*y)
    status = model.optimize(mipx.MAXIMIZE)
    if status == mipx.OPTIMAL:
        print("Optimal solution found")
        print("Obj value:", model.ObjVal)
        mipx.debugVar(x)
        mipx.debugVar(y)
        model.reportMultiObjValue()
    else:
        print("No solution found")



#         .__
#   _____ |__|________  ___
#  /     \|  \____ \  \/  /
# |  Y Y  \  |  |_> >    <
# |__|_|  /__|   __/__/\_ \
#       \/   |__|        \/
__title__ = "mipx"
__description__ = "Python (Mixed-Integer Linear Programming Constraint Programming) Optimization Tools"
__version__ = "0.5.11"
__author__ = "luyi"
__author_email__ = "2662017230@qq.com"

__copyright__ = "Copyright 2024 luyi"
